"""Test db importing module."""

import pytest

from py_research.db import DB
from py_research.db.importing import TableMap, tree_to_db


@pytest.fixture
def nested_db_dict() -> dict:
    """Return nested data dict for import testing."""
    return {
        "resultCount": 3,
        "search": "test",
        "results": [
            {
                "project_name": "baking cake",
                "project_start": "2020-01-01",
                "project_end": "2020-01-04",
                "project_status": "done",
                "organization_name": "Bakery",
                "organization_address": "Main Street 1",
                "organization_city": "Bakerville",
                "tasks": [
                    {
                        "task_name": "task1",
                        "task_assignee": "John",
                        "task_status": "todo",
                    },
                    {
                        "task_name": "task2",
                        "task_assignee": "John",
                        "task_status": "todo",
                    },
                    {
                        "task_name": "task3",
                        "task_assignee": "Jane",
                        "task_status": "done",
                    },
                ],
                "members": [
                    {"name": "John", "age": 20, "role": "baker"},
                    {"name": "John", "age": 20, "role": "manager"},
                ],
            },
            {
                "project_name": "cleaning shoes",
                "project_start": "2020-01-02",
                "project_end": "2020-01-05",
                "project_status": "done",
                "organization_name": "Shoe Shop",
                "organization_address": "Main Street 2",
                "organization_city": "Shoetown",
                "tasks": [
                    {
                        "task_name": "task4",
                        "task_assignee": "John",
                        "task_status": "todo",
                    },
                    {
                        "task_name": "task5",
                        "task_assignee": "Jane",
                        "task_status": "todo",
                    },
                ],
                "members": [
                    {"name": "John", "age": 20, "role": "cleaner"},
                    {"name": "Jane", "age": 30, "role": "manager"},
                ],
            },
            {
                "project_name": "fixing cars",
                "project_start": "2020-01-03",
                "project_end": "2020-01-06",
                "project_status": "started",
                "organization_name": "Car Shop",
                "organization_address": "Main Street 3",
                "organization_city": "Cartown",
                "tasks": [
                    {
                        "task_name": "task6",
                        "task_assignee": "John",
                        "task_status": "todo",
                    },
                ],
                "members": [
                    {"name": "John", "age": 20, "role": "mechanic"},
                    {"name": "Jane", "age": 30, "role": "manager"},
                    {"name": "Jack", "age": 40, "role": "manager"},
                ],
            },
        ],
    }


@pytest.fixture
def root_table_mapping() -> TableMap:
    """Return root table mapping for import testing."""
    return TableMap(
        table="searches",
        id_type="hash",
        id_attr="search_term",
        map={
            "resultCount": "result_count",
            "search": "search_term",
            "results": TableMap(
                table="projects",
                id_type="uuid",
                map={
                    "project_name": "name",
                    "project_start": "start",
                    "project_end": "end",
                    "project_status": "status",
                    "tasks": TableMap(
                        table="tasks",
                        map={
                            "task_name": "name",
                            "task_assignee": TableMap(
                                table="users",
                                map="name",
                                match_by_attr="name",
                            ),
                            "task_status": "status",
                        },
                        id_type="uuid",
                    ),
                    "members": TableMap(
                        table="users",
                        map={"name", "age"},
                        join_table_name="memberships",
                        join_table_map={
                            "role": "role",
                        },
                    ),
                },
                ext_maps=[
                    TableMap(
                        table="organizations",
                        map={
                            "organization_name": "name",
                            "organization_address": "address",
                            "organization_city": "city",
                        },
                        link_type="n-1",
                        link_attr="organization",
                        id_type="attr",
                        id_attr="name",
                    ),
                ],
            ),
        },
    )


def test_import_db_from_tree(nested_db_dict: dict, root_table_mapping: TableMap):
    """Test importing nested data dict to database."""
    db = tree_to_db(nested_db_dict, root_table_mapping, False)
    assert isinstance(db, DB)
    assert set(db.keys()) == {
        "searches",
        "projects",
        "tasks",
        "users",
        "organizations",
        "memberships",
        "searches_projects",
        "projects_tasks",
    }
    assert len(db["searches"].df) == 1
    assert len(db["projects"].df) == 3
    assert len(db["tasks"].df) == 6
    assert len(db["users"].df) == 3
    assert len(db["organizations"].df) == 3
    assert len(db["memberships"].df) > 0
    assert len(db["searches_projects"].df) > 0
    assert len(db["projects_tasks"].df) > 0
    assert db.relations == {
        ("searches_projects", "results"): ("projects", "_id"),
        ("searches_projects", "results_of"): ("searches", "_id"),
        ("projects_tasks", "tasks"): ("tasks", "_id"),
        ("projects_tasks", "tasks_of"): ("projects", "_id"),
        ("memberships", "members"): ("users", "_id"),
        ("memberships", "members_of"): ("projects", "_id"),
        ("projects", "organization"): ("organizations", "_id"),
    }
    assert db.join_tables == {"searches_projects", "projects_tasks", "memberships"}



"""Test the colors module."""

import pytest
from py_research.colors import ColorTheme, get_theme, to_bg_color


def test_get_theme():
    """Test the get_theme function."""
    theme = get_theme()
    assert isinstance(theme, ColorTheme)


def test_to_bg_color():
    """Test the to_bg_color function."""
    color = "#FF0000"  # red
    lightness = 0.8
    bg_color = to_bg_color(color, lightness)
    assert isinstance(bg_color, str)
    assert bg_color.startswith("#")

    # Test with invalid color
    with pytest.raises(ValueError):
        to_bg_color("invalid_color", lightness)



"""Base classes and types for relational database package."""

from collections.abc import Hashable, Iterable, Sequence
from dataclasses import asdict, dataclass, field
from functools import reduce
from pathlib import Path
from typing import Any, Literal, cast, overload

import pandas as pd

from py_research.data import parse_dtype
from py_research.reflect import PyObjectRef

from .conflicts import DataConflictError, DataConflictPolicy


def _unmerge(df: pd.DataFrame) -> dict[str, pd.DataFrame]:
    """Extract this table into its own database.

    Returns:
        Database containing only this table.
    """
    if df.columns.nlevels != 2:
        raise ValueError("Cannot only unmerge dataframe with two column levels.")

    return {
        str(s): cast(pd.DataFrame, df[s]).drop_duplicates()
        for s in df.columns.get_level_values(0).unique()
    }


@dataclass
class Table:
    """Table in a relational database."""

    # Attributes:

    db: "DB"
    """Database this table belongs to."""

    df: pd.DataFrame
    """Dataframe containing the data of this table."""

    source_map: str | dict[str, str]
    """Mapping to source tables of this table.

    For single source tables, this is a string containing the name of the source table.

    For multiple source tables, the dataframe hast multi-level columns. ``source_map``
    is then a mapping from the first level of these columns to the source tables.
    """

    indexes: dict[str, str] = field(default_factory=dict)
    """Mapping from source table names to index column names."""

    # Creation:

    @staticmethod
    def from_excel(path: Path) -> "Table":
        """Load relational table from Excel file."""
        source_map = pd.read_excel(path, sheet_name="_source_tables", index_col=0)[
            "table"
        ].to_dict()

        indexes = pd.read_excel(path, sheet_name="_indexes", index_col=0)[
            "column"
        ].to_dict()

        data = pd.read_excel(
            path,
            sheet_name="data",
            index_col=0,
            header=(0 if len(source_map) == 1 else (0, 1)),
        )

        df_dict = (
            {source_map[s]: df for s, df in _unmerge(data).items()}
            if len(source_map) > 1
            else {list(source_map.keys())[0]: data}
        )

        return Table(
            DB(df_dict),
            data,
            source_map if len(source_map) > 1 else list(source_map.keys())[0],
            indexes,
        )

    # Public methods:

    def to_excel(self, path: Path) -> None:
        """Save this table to an Excel file."""
        sheets = {
            "_source": pd.Series({"database": self.db.backend})
            .rename("value")
            .to_frame(),
            "_source_tables": pd.DataFrame.from_dict(
                {k: [v] for k, v in self.source_map.items()}
                if isinstance(self.source_map, dict)
                else {self.source_map: [None]},
                orient="index",
                columns=["table"],
            ).rename_axis(index="col_prefix"),
            "_indexes": pd.DataFrame.from_dict(
                {k: [v] for k, v in self.indexes.items()},
                orient="index",
                columns=["column"],
            ).rename_axis(index="table"),
            "data": self.df,
        }

        with pd.ExcelWriter(
            path,
            engine="openpyxl",
        ) as writer:
            for name, df in sheets.items():
                df.to_excel(writer, sheet_name=name, index=True)

    def filter(self, filter_series: pd.Series) -> "Table":
        """Filter this table.

        Args:
            filter_series: Boolean series to filter the table with.

        Returns:
            New table containing the filtered data.
        """
        return Table(
            self.db,
            self.df.loc[filter_series],
            self.source_map,
        )

    @overload
    def merge(
        self,
        right: "SingleTable | None" = None,
        link_to_right: str | tuple[str, str] = ...,
        link_to_left: str | None = None,
        link_table: "SingleTable | None" = None,
        naming: Literal["source", "path"] = ...,
    ) -> "Table":
        ...

    @overload
    def merge(
        self,
        right: "SingleTable" = ...,
        link_to_right: str | tuple[str, str] | None = None,
        link_to_left: str | None = None,
        link_table: "SingleTable | None" = None,
        naming: Literal["source", "path"] = ...,
    ) -> "Table":
        ...

    @overload
    def merge(
        self,
        right: "SingleTable | None" = ...,
        link_to_right: str | tuple[str, str] | None = None,
        link_to_left: str | None = None,
        link_table: "SingleTable" = ...,
        naming: Literal["source", "path"] = ...,
    ) -> "Table":
        ...

    def merge(  # noqa: C901
        self,
        right: "SingleTable | None" = None,
        link_to_right: str | tuple[str, str] | None = None,
        link_to_left: str | None = None,
        link_table: "SingleTable | None" = None,
        naming: Literal["source", "path"] = "source",
    ) -> "Table":
        """Merge this table with another, returning a new table.

        Args:
            link_to_right: Name of column to use for linking from left to right table.
            link_to_left: Name of column to use for linking from right to left table.
            right: Other (left) table to merge with.
            link_table: Link table (join table) to use for double merging.
            naming:
                Naming strategy to use for naming the first level of merged columns.
                Use "path" if you merge multiple times from the same source table.

        Note:
            At least one of ``link_to_right``, ``right`` or ``link_table`` must be
            supplied.

        Returns:
            New table containing the merged data.
            The returned table will have a multi-level column index,
            where the first level references the source table of each column
            via the ``source_map`` attribute.
        """
        # Standardize format of dataframe, making sure its columns are multi-level.
        merged = self.df
        if isinstance(self.source_map, str):
            merged = merged.rename_axis(merged.index.name or "id", axis="index")
            merged = merged.reset_index()
            merged.columns = pd.MultiIndex.from_product(
                [[self.source_map], merged.columns]
            )

        # Standardize format of source map, making sure it is a dict.
        merge_source_map = (
            self.source_map
            if isinstance(self.source_map, dict)
            else {self.source_map: self.source_map}
        )
        inv_src_map = {
            v: [k for k, v2 in merge_source_map.items() if v2 == v]
            for v in set(merge_source_map.values())
        }
        sources = set(merge_source_map.values())

        merge_indexes = self.indexes.copy()

        # Set up a list of all merges to be applied with their parameters.
        merges: list[tuple[tuple[str, str], str, tuple[SingleTable, str]]] = []

        # Get a list of all applicable forward merges with their parameters.
        if link_to_right is not None:
            rels = (
                [
                    ((sp, link_to_right), self.db.relations.get((s, link_to_right)))
                    for s in sources
                    for sp in inv_src_map[s]
                ]
                if isinstance(link_to_right, str)
                else [(link_to_right, self.db.relations.get(link_to_right))]
            )
            rels = [(s, r) for s, r in rels if r is not None]
            if len(rels) == 0:
                raise ValueError(
                    "Cannot find relation with source table "
                    f"in {sources} and source col = '{link_to_right}'."
                )

            if right is not None:
                rels = [(s, r) for s, r in rels if r[0] == right.name]
                if len(rels) == 0:
                    raise ValueError(
                        f"Cannot find relation with source table in {sources}, "
                        f"source col = '{link_to_right}' and "
                        f"target table = {right.name}."
                    )

            merges += [
                (
                    (sp, sc),
                    tt
                    if naming == "source"
                    else f"{sp}->{sc}=>{tc}<-{tt}"
                    if tc != (self.db[tt].df.index.name or "id")
                    else f"{sp}->{sc}",
                    (right or self.db[tt], tc),
                )
                for (sp, sc), (tt, tc) in rels
            ]
        elif right is None and link_table is None:
            outgoing = self.db._get_rels(
                sources=list(self.source_map.values())
                if isinstance(self.source_map, dict)
                else [self.source_map]
            )
            merges += [
                (
                    (sp, sc),
                    tt
                    if naming == "source"
                    else f"{sp}->{sc}=>{tc}<-{tt}"
                    if tc != (self.db[tt].df.index.name or "id")
                    else f"{sp}->{sc}",
                    (self.db[tt], tc),
                )
                for st, sc, tt, tc in outgoing.itertuples(index=False)
                for sp in inv_src_map[st]
                if right is None or right.name == tt
            ]

        # Get all incoming relations.
        incoming = self.db._get_rels(
            targets=list(self.source_map.values())
            if isinstance(self.source_map, dict)
            else [self.source_map]
        )

        # Get a list of all applicable backward merges with their parameters.
        if len(merges) == 0 and right is not None:
            # Get all matchin incoming links.
            from_right = [
                (st, sc, tt, tp, tc)
                for st, sc, tt, tc in incoming.itertuples(index=False)
                for tp in inv_src_map[tt]
                if right.name == st
            ]

            # If there is only one, use it as explicit link_to_left.
            if len(from_right) == 1:
                link_to_left = from_right[0][1]

            if link_to_left is not None:
                rel = self.db.relations.get((right.name, link_to_left))
                if rel is None:
                    raise ValueError(
                        f"Cannot find relation with target table in {self.source_map} "
                        f", src table = {right.name} and src col = '{link_to_left}'."
                    )
                tt, tc = rel
                merges += [
                    (
                        (tp, tc),
                        tt
                        if naming == "source"
                        else f"{tp}->{tc}<={right.name}"
                        if tc != (self.db[tt].df.index.name or "id")
                        else f"{tp}<={right.name}",
                        (right, link_to_left),
                    )
                    for tp in inv_src_map[tt]
                ]
            else:
                merges += [
                    (
                        (tp, tc),
                        tt
                        if naming == "source"
                        else f"{tp}->{tc}<={sc}<-{st}"
                        if tc != (self.db[tt].df.index.name or "id")
                        else f"{tp}<={sc}<-{st}",
                        (self.db[st], sc),
                    )
                    for st, sc, tt, tp, tc in from_right
                ]

        # Get a list of all related join / link tables.
        link_tables = []
        if link_table is not None:
            link_tables = [link_table]
        else:
            link_tables = [
                self.db[t]
                for t in incoming["source_table"].unique()
                if t in self.db.join_tables
            ]

        # Get a list of all applicable double merges with their parameters.
        if len(merges) == 0:
            for jt in link_tables:
                jt_links = self.db._get_rels(
                    sources=list(jt.source_map.values())
                    if isinstance(jt.source_map, dict)
                    else [jt.source_map]
                )
                backlinks = jt_links.loc[jt_links["target_table"].isin(sources)]
                other_links = jt_links.loc[~jt_links["target_table"].isin(sources)]

                if right is not None:
                    other_links = other_links.loc[
                        other_links["target_table"] == right.name
                    ]
                    if len(other_links) == 0:
                        continue

                for _, bsc, btt, btc in backlinks.itertuples(index=False):
                    backlink_count = sum(backlinks["target_table"] == btt)
                    for btp in inv_src_map[btt]:
                        jt_prefix = (
                            jt.name
                            if naming == "source"
                            else (
                                (
                                    f"{btp}->{btc}"
                                    if btc != (self.db[btt].df.index.name or "id")
                                    else btp
                                )
                                + "<="
                                + (
                                    f"{bsc}<-{jt.name}"
                                    if backlink_count > 1
                                    else jt.name
                                )
                            )
                        )
                        # First perform a merge with the joint table.
                        merges.append(
                            (
                                (btp, btc),
                                jt_prefix,
                                (jt, bsc),
                            )
                        )
                        # Then perform a merge with all other tables linked from there.
                        merges += [
                            (
                                (jt_prefix, osc),
                                ott
                                if naming == "source"
                                else (
                                    (jt_prefix if len(link_tables) > 1 else btp)
                                    + "->"
                                    + (
                                        f"{osc}<={otc}<-{ott}"
                                        if otc != (self.db[ott].df.index.name or "id")
                                        else osc
                                    )
                                ),
                                (right or self.db[ott], otc),
                            )
                            for _, osc, ott, otc in other_links.itertuples(index=False)
                        ]

        if len(merges) == 0:
            raise ValueError(
                "Could not find any relations between the given tables to merge on."
            )

        for (sp, sc), tp, (tt, tc) in merges:
            # Flatten multi-level columns of left table.
            left_df = merged.copy()
            left_df.columns = merged.columns.map(lambda c: "->".join(c))
            left_on = f"{sp}->{sc}"

            # Properly name columns of right table.
            right_df = tt.df.reset_index().rename(columns=lambda c: f"{tp}->{c}")
            right_on = f"{tp}->{tc}"

            new_merged = left_df.merge(
                right_df,
                left_on=left_on,
                right_on=right_on,
                how="left",
            )

            # Restore two-level column index.
            new_merged.columns = pd.MultiIndex.from_tuples(
                [
                    ("->".join(c[:-1]), c[-1])
                    for c in new_merged.columns.map(
                        lambda c: str(c).split("->")
                    ).to_list()
                ]
            )
            merged = new_merged.copy()

            # Add to source_map
            merge_source_map[tp] = tt.name

            # Add to indexes
            merge_indexes[tt.name] = tt.df.index.name or "id"

        return Table(self.db, merged, merge_source_map, merge_indexes)

    def flatten(
        self,
        sep: str = ".",
        prefix_strategy: Literal["always", "on_conflict"] = "always",
    ) -> pd.DataFrame:
        """Collapse multi-dim. column labels of multi-source table, returning new df.

        Args:
            sep: Separator to use between column levels.
            prefix_strategy: Strategy to use for prefixing column names.

        Returns:
            Dataframe representation of this table with flattened multi-dim columns.
        """
        level_counts = (
            self.df.columns.to_frame()
            .groupby(level=(1 if self.df.columns.nlevels > 1 else 0))
            .size()
        )

        res_df = self.df.copy()
        res_df.columns = [
            c[0]
            if len(c) == 1
            else c[1]
            if (
                len(c) > 1
                and prefix_strategy == "on_conflict"
                and level_counts[c[1]] == 1
            )
            else sep.join(c)
            for c in self.df.columns.to_frame().itertuples(index=False)
        ]

        return res_df

    def extract(self, with_relations: bool = True) -> "DB":
        """Extract this table into its own database.

        Returns:
            Database containing only this table.
        """
        source_map = (
            self.source_map
            if isinstance(self.source_map, dict)
            else {self.source_map: self.source_map}
        )
        inv_src_map = {
            v: [k for k, v2 in source_map.items() if v2 == v]
            for v in set(source_map.values())
        }

        source_tables = {
            t: pd.concat(
                [cast(pd.DataFrame, self.df[s]).drop_duplicates() for s in all_s]
            )
            .dropna(subset=[self.indexes[t]])
            .set_index(self.indexes[t])
            for t, all_s in inv_src_map.items()
        }

        return (
            self.db.filter(
                {
                    s: self.db[s].df.index.to_series().isin(df.index)
                    for s, df in source_tables.items()
                }
            )
            if with_relations
            else DB(table_dfs=source_tables)
        )

    # Dictionary interface:

    def __getitem__(self, name: str) -> pd.Series:  # noqa: D105
        return self.df[name]

    def __contains__(self, name: str) -> bool:  # noqa: D105
        return name in self.df.columns

    def __iter__(self) -> Iterable[Hashable]:  # noqa: D105
        return iter(self.df)

    def __len__(self) -> int:  # noqa: D105
        return len(self.df)

    def keys(self) -> Iterable[str]:  # noqa: D102
        return self.df.keys()

    def get(self, name: str, default: Any = None) -> Any:  # noqa: D102
        return self.df.get(name, default)

    # DataFrame interface:

    @property
    def columns(self) -> Sequence[str | tuple[str, str]]:
        """Return the columns of this table."""
        return self.df.columns.tolist()


class SingleTable(Table):
    """Relational database table with a single source table."""

    # Custom constructor:

    def __init__(
        self,
        name: str,
        db: "DB",
        df: pd.DataFrame,
    ) -> None:
        """Initialize this table.

        Args:
            name: Name of the source table.
            db: Database this table belongs to.
            df: Dataframe containing the data of this table.
        """
        self.name = name
        self.db = db
        self.df = df

    # Computed properties:

    @property
    def source_map(self) -> str:  # type: ignore[override]
        """Name of the source table of this table."""
        return self.name

    @property
    def indexes(self) -> dict[str, str]:  # type: ignore[override]
        """Name of the source table of this table."""
        return {self.name: self.df.index.name or "id"}

    # Method overrides:

    def filter(self, filter_series: pd.Series) -> "SingleTable":
        """Return a filtered version of this table."""
        table = super().filter(filter_series)
        assert isinstance(table.source_map, str)
        return SingleTable(name=table.source_map, db=table.db, df=table.df)

    def trim(self, cols: list[str]) -> "SingleTable":
        """Return a trimmed version of this table.

        Args:
            cols: Columns to keep.
        """
        return SingleTable(name=self.name, db=self.db, df=self.df[cols])

    # Public methods:

    def extend(
        self,
        other: "pd.DataFrame",
        conflict_policy: DataConflictPolicy | dict[str, DataConflictPolicy] = "raise",
    ) -> "SingleTable":
        """Extend this table with data from another, returning a new table.

        Args:
            other: Other table to extend with.
            conflict_policy:
                Policy to use for resolving conflicts. Can be a global setting,
                per-column via supplying a dict with column names as keys.

        Returns:
            New table containing the extended data.
        """
        # Align index and columns of both tables.
        left, right = self.df.align(other)

        # First merge data, ignoring conflicts per default.
        extended_df = left.combine_first(right)

        # Find conflicts, i.e. same-index & same-column values
        # that are different in both tables and neither of them is NaN.
        conflicts = ~((left == right) | left.isna() | right.isna())

        if any(conflicts):
            # Deal with conflicts according to `conflict_policy`.

            # Determine default policy.
            default_policy: DataConflictPolicy = (
                conflict_policy if isinstance(conflict_policy, str) else "raise"
            )
            # Expand default policy to all columns.
            policies: dict[str, DataConflictPolicy] = {
                c: default_policy for c in conflicts.columns
            }
            # Assign column-level custom policies.
            if isinstance(conflict_policy, dict):
                policies = {**policies, **conflict_policy}

            # Iterate over columns and associated policies:
            for c, p in policies.items():
                # Only do conflict resolution if there are any for this col.
                if any(conflicts[c]):
                    match p:
                        case "override":
                            # Override conflicts with data from left.
                            extended_df[c][conflicts[c]] = right[conflicts[c]]
                            extended_df[c] = extended_df[c]

                        case "ignore":
                            # Nothing to do here.
                            pass
                        case "raise":
                            # Raise an error.
                            raise DataConflictError(
                                {
                                    (c, c, str(i)): (lv, rv)
                                    for (i, lv), (i, rv) in zip(
                                        left.loc[conflicts[c]][c].items(),
                                        right.loc[conflicts[c]][c].items(),
                                    )
                                }
                            )
                        case _:
                            pass

        return SingleTable(
            self.name,
            self.db,
            extended_df,
        )


class SourceTable(SingleTable):
    """Original table in a relational database."""

    # Custom constuctor:

    def __init__(self, name: str, db: "DB") -> None:
        """Initialize this table.

        Args:
            name: Name of the source table.
            db: Database this table belongs to.
        """
        self.name = name
        self.db = db

    # Computed properties:

    @property
    def df(self) -> pd.DataFrame:
        """Return the dataframe of this table."""
        return self.db.table_dfs[self.name]


class DBSchema:
    """Base class for static database schemas."""


@dataclass
class DB:
    """Relational database consisting of multiple named tables."""

    # Attributes:

    table_dfs: dict[str, pd.DataFrame] = field(default_factory=dict)
    """Dataframes containing the data of each table in this database."""

    relations: dict[tuple[str, str], tuple[str, str]] = field(default_factory=dict)
    """Relations between tables in this database."""

    join_tables: set[str] = field(default_factory=set)
    """Names of tables that are used as n-to-m join tables in this database."""

    schema: type[DBSchema] | None = None
    """Schema of this database."""

    updates: dict[pd.Timestamp, dict[str, Any]] = field(default_factory=dict)
    """List of update times with comments, tags, authors, etc."""

    backend: Path | None = None
    """File backend of this database, hence where it was loaded from and is
    saved to by default.
    """

    _copied: bool = False
    """Whether this database has been copied from somewhere else."""

    # Construction:

    def __post_init__(self) -> None:  # noqa: D105
        if not self._copied:
            self.updates[pd.Timestamp.now().round("s")] = {
                "action": "construct",
                "table_dfs": str(set(self.table_dfs.keys())),
                "relations": str(set(self.relations.items())),
                "join_tables": str(self.join_tables),
            }

    # Public methods:

    @staticmethod
    def load(path: Path | str, auto_parse_dtype: bool = False) -> "DB":
        """Load a database from an excel file.

        Args:
            path: Path to the excel file.
            auto_parse_dtype: Whether to automatically parse the dtypes of the data.

        Returns:
            Database object.
        """
        path = Path(path) if isinstance(path, str) else path

        relations = {}
        if "_relations" in pd.ExcelFile(path).sheet_names:
            rel_df = pd.read_excel(path, sheet_name="_relations", index_col=[0, 1])
            relations = rel_df.apply(
                lambda r: (r["target_table"], r["target_col"]),
                axis="columns",
            ).to_dict()

        join_tables = set()
        if "_join_tables" in pd.ExcelFile(path).sheet_names:
            jt_df = pd.read_excel(path, sheet_name="_join_tables", index_col=0)
            join_tables = set(jt_df["name"].tolist())

        schema = None
        if "_schema" in pd.ExcelFile(path).sheet_names:
            schema_df = pd.read_excel(path, sheet_name="_schema", index_col=0)
            schema_ref = PyObjectRef(**schema_df["value"].to_dict())
            schema_ref.type = type
            if pd.isna(schema_ref.version):
                schema_ref.version = None

            schema = schema_ref.resolve()
            if not issubclass(schema, DBSchema):
                raise ValueError("Database schema must be a subclass of `DBSchema`.")

        updates = {}
        if "_updates" in pd.ExcelFile(path).sheet_names:
            update_df = pd.read_excel(path, sheet_name="_updates", index_col=0).drop(
                columns=["comment"]
            )
            update_df.index = pd.to_datetime(
                update_df.index, infer_datetime_format=True
            )
            updates = {
                cast(pd.Timestamp, t): c.to_dict() for t, c in update_df.iterrows()
            }

        df_dict = {
            str(k): df.apply(parse_dtype, axis="index") if auto_parse_dtype else df
            for k, df in pd.read_excel(path, sheet_name=None, index_col=0).items()
            if not k.startswith("_")
        }

        return DB(df_dict, relations, join_tables, schema, updates, path, True)

    def save(self, path: Path | str | None = None) -> None:
        """Save this database to an excel file.

        Args:
            path:
                Path to excel file. Will be overridden if it exists.
                Uses this database's backend as default path, if none given.
        """
        path = Path(path) if isinstance(path, str) else path
        if path is None:
            if self.backend is None:
                raise ValueError(
                    "Need to supply explicit path for databases without a backend."
                )
            path = self.backend

        sheets = {
            **(
                {
                    "_schema": pd.Series(asdict(PyObjectRef.reference(self.schema)))
                    .rename("value")
                    .to_frame()
                    .rename_axis(index="key")
                }
                if self.schema is not None
                else {}
            ),
            "_relations": pd.DataFrame.from_records(
                list(self.relations.values()),
                columns=["target_table", "target_col"],
                index=pd.MultiIndex.from_tuples(list(self.relations.keys())),
            ).rename_axis(index=["source_table", "source_col"]),
            "_join_tables": pd.DataFrame({"name": list(self.join_tables)}),
            "_updates": pd.DataFrame.from_dict(
                {t.strftime("%Y-%m-%d %X"): d for t, d in self.updates.items()},
                orient="index",
            )
            .rename_axis(index="time")
            .assign(comment=""),
            **self.table_dfs,
        }
        with pd.ExcelWriter(
            path,
            engine="openpyxl",
        ) as writer:
            for name, sheet in sheets.items():
                sheet.to_excel(writer, sheet_name=name, index=True)

    def describe(self) -> dict[str, dict[str, str]]:
        """Return a description of this database.

        Returns:
            Mapping of table names to table descriptions.
        """
        join_tables = {
            name: f"{len(self[name])} links"
            + (
                f" x {n_attrs} attributes"
                if (n_attrs := len(self[name].columns) - 2) > 0
                else ""
            )
            for name in self.join_tables
        }

        return {
            "tables": {
                name: f"{len(df)} objects x {len(df.columns)} attributes"
                for name, df in self.items()
                if name not in join_tables
            },
            "join tables": join_tables,
        }

    def copy(self, deep: bool = True) -> "DB":
        """Create a copy of this database, optionally deep.

        Args:
            deep: Whether to perform a deep copy (copy all dataframes).

        Returns:
            Copy of this database.
        """
        return DB(
            table_dfs={
                name: (table.df.copy() if deep else table.df)
                for name, table in self.items()
            },
            relations=self.relations,
            join_tables=self.join_tables,
            schema=self.schema,
            updates=self.updates,
            _copied=True,
        )

    def extend(
        self,
        other: "DB | dict[str, pd.DataFrame] | Table",
        conflict_policy: DataConflictPolicy
        | dict[str, DataConflictPolicy | dict[str, DataConflictPolicy]] = "raise",
    ) -> "DB":
        """Extend this database with data from another, returning a new database.

        Args:
            other: Other database, dataframe dict or table to extend with.
            conflict_policy:
                Policy to use for resolving conflicts. Can be a global setting,
                per-table via supplying a dict with table names as keys, or per-column
                via supplying a dict of dicts.

        Returns:
            New database containing the extended data.
        """
        # Standardize type of other.
        other = (
            other
            if isinstance(other, DB)
            else DB(other)
            if isinstance(other, dict)
            else other.extract()
        )

        # Get the union of all table (names) in both databases.
        tables = set(self.keys()) | set(other.keys())

        # Set up variable to contain the merged database.
        merged: dict[str, pd.DataFrame] = {}

        for t in tables:
            if t not in self:
                merged[t] = other[t].df if isinstance(other, DB) else other[t]
            elif t not in other:
                merged[t] = self[t].df
            # Perform more complicated matching if table exists in both databases.
            else:
                table_conflict_policy = (
                    conflict_policy
                    if isinstance(conflict_policy, str)
                    else conflict_policy.get(t) or "raise"
                )
                merged[t] = (
                    self[t]
                    .extend(
                        other[t].df if isinstance(other, DB) else other[t],
                        conflict_policy=table_conflict_policy,
                    )
                    .df
                )

        return DB(
            table_dfs=merged,
            relations={**self.relations, **other.relations},
            join_tables={*self.join_tables, *other.join_tables},
            schema=self.schema,
            updates={
                **self.updates,
                pd.Timestamp.now().round("s"): {
                    "action": "extend",
                    "table_dfs": str(set(other.table_dfs.keys())),
                    "relations": str(set(other.relations.items())),
                    "join_tables": str(other.join_tables),
                },
            },
        )

    def trim(
        self,
        centers: list[str] | None = None,
        circuit_breakers: list[str] | None = None,
    ) -> "DB":
        """Return new database minus orphan data (relative to given ``centers``).

        Args:
            centers: Tables to use as centers for the trim.
            circuit_breakers: Tables to use as circuit breakers for the trim.

        Returns:
            New database containing the trimmed data.
        """
        res = (
            self._isotropic_trim()
            if centers is None
            else self._centered_trim(centers, circuit_breakers)
        )
        res.updates[pd.Timestamp.now().round("s")] = {
            "action": "trim",
            "centers": str(set(centers or [])),
            "circuit_breakers": str(set(circuit_breakers or [])),
            "remaining_tables": str(set(self.keys())),
        }
        return res

    def filter(
        self, filters: dict[str, pd.Series], extra_cb: list[str] | None = None
    ) -> "DB":
        """Return new db only containing data related to rows matched by ``filters``.

        Args:
            filters: Mapping of table names to boolean filter series
            extra_cb:
                Additional circuit breakers (on top of the filtered tables)
                to use when trimming the database according to the filters

        Returns:
            New database containing the filtered data.
            The returned database will only contain the filtered tables and
            all tables that have (indirect) references to them.

        Note:
            This is equivalent to trimming the database with the filtered tables
            as centers and the filtered tables and ``extra_cb`` as circuit breakers.
        """
        # Filter out unmatched rows of filter tables.
        new_db = DB(
            table_dfs={
                t: (table.df[filters[t]] if t in filters else table.df)
                for t, table in self.items()
            },
            relations=self.relations,
            join_tables=self.join_tables,
        )

        # Always use the filter tables as circuit_breakes.
        # Otherwise filtered-out rows may be re-included.
        cb = list(set(filters.keys()) | set(extra_cb or []))

        # Trim all other tables such that only rows with (indirect) references to
        # remaining rows in filter tables are left.
        res = new_db.trim(list(filters.keys()), circuit_breakers=cb)
        res.updates[pd.Timestamp.now().round("s")] = {
            "action": "filter",
            "filter_tables": str(set(filters.keys())),
            "extra_cb": str(set(extra_cb or [])),
            "remaining_tables": str(set(self.keys())),
        }
        return res

    def to_graph(
        self, nodes: Sequence[Table | str]
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Export links between select database objects in a graph format.

        E.g. for usage with `Gephi`_

        .. _Gephi: https://gephi.org/
        """
        nodes = [self[n] if isinstance(n, str) else n for n in nodes]
        # Concat all node tables into one.
        node_dfs = [
            n.df.reset_index().assign(table=n.source_map)
            if isinstance(n.source_map, str)
            else pd.concat(
                [
                    cast(pd.DataFrame, n.df[p])
                    .drop_duplicates()
                    .reset_index()
                    .assign(table=s)
                    for p, s in n.source_map.items()
                    if s not in self.join_tables
                ],
                ignore_index=True,
            )
            for n in nodes
        ]
        node_df = (
            pd.concat(node_dfs, ignore_index=True)
            .reset_index()
            .rename(columns={"index": "node_id"})
        )

        if "level_0" in node_df.columns:
            node_df = node_df.drop(columns=["level_0"])

        # Find all link tables between the given node tables.
        node_names = list(node_df["table"].unique())

        directed_edges = self._get_rels(sources=node_names, targets=node_names)
        undirected_edges = pd.concat(
            [self._get_rels(sources=[j], targets=node_names) for j in self.join_tables]
        )

        # Concat all edges into one table.
        edge_df = pd.concat(
            [
                *[
                    node_df.loc[node_df["table"] == str(rel["source_table"])]
                    .rename(columns={"node_id": "source"})
                    .merge(
                        node_df.loc[node_df["table"] == str(rel["target_table"])],
                        left_on=rel["source_col"],
                        right_on=rel["target_col"],
                    )
                    .rename(columns={"node_id": "target"})[["source", "target"]]
                    .assign(ltr=rel["source_col"], rtl=None)
                    for _, rel in directed_edges.iterrows()
                ],
                *[
                    self[str(source_table)]
                    .df.merge(
                        node_df.loc[
                            node_df["table"] == str(rel.iloc[0]["target_table"])
                        ].dropna(axis="columns", how="all"),
                        left_on=rel.iloc[0]["source_col"],
                        right_on=rel.iloc[0]["target_col"],
                        how="inner",
                    )
                    .rename(columns={"node_id": "source"})
                    .merge(
                        node_df.loc[
                            node_df["table"] == str(rel.iloc[1]["target_table"])
                        ].dropna(axis="columns", how="all"),
                        left_on=rel.iloc[1]["source_col"],
                        right_on=rel.iloc[1]["target_col"],
                        how="inner",
                    )
                    .rename(columns={"node_id": "target"})[
                        list(
                            {"source", "target", *self[str(source_table)].columns}
                            - {rel.iloc[0]["source_col"], rel.iloc[1]["source_col"]}
                        )
                    ]
                    .assign(
                        ltr=rel.iloc[1]["source_col"], rtl=rel.iloc[0]["source_col"]
                    )
                    for source_table, rel in undirected_edges.groupby("source_table")
                    if len(rel) == 2  # We do not export hyper-graphs.
                ],
            ],
            ignore_index=True,
        )

        return node_df, edge_df

    # Dictionary interface:

    def __getitem__(self, name: str) -> SingleTable:  # noqa: D105
        if name.startswith("_"):
            raise KeyError(f"Table '{name}' does not exist in this database.")
        return (
            SourceTable(name=name, db=self)
            if name in self.table_dfs
            else self._manifest_virtual_table(name)
        )

    def __setitem__(  # noqa: D105
        self, name: str, value: SingleTable | pd.DataFrame
    ) -> None:
        if name.startswith("_"):
            raise KeyError("Table names starting with '_' are not allowed.")

        value = value.df if isinstance(value, SingleTable) else value
        rels = self._get_rels()
        target_tables = rels["target_table"].unique()
        if name in target_tables:
            target_cols = rels.loc[rels["target_table"] == name, "target_col"].unique()
            if not set(target_cols).issubset(value.reset_index().columns):
                raise ValueError(
                    "Relations to given table name already exist, "
                    f"but not all referenced columns were supplied: {target_cols}."
                )

        self.table_dfs[name] = value
        self.updates[pd.Timestamp.now().round("s")] = {
            "action": "set_table",
            "table": name,
        }

    def __contains__(self, name: str) -> bool:  # noqa: D105
        return name in self.keys() and not name.startswith("_")

    def __iter__(self) -> Iterable[str]:  # noqa: D105
        return (k for k in self.keys() if not k.startswith("_"))

    def __len__(self) -> int:  # noqa: D105
        return len(list(k for k in self.keys() if not k.startswith("_")))

    def keys(self) -> set[str]:  # noqa: D102
        return set(k for k in self.table_dfs if not k.startswith("_")) | set(
            self._get_rels()["target_table"].unique()
        )

    def values(self) -> Iterable[SingleTable]:  # noqa: D102
        return [self[name] for name in self.keys() if not name.startswith("_")]

    def items(self) -> Iterable[tuple[str, SingleTable]]:  # noqa: D102
        return [(name, self[name]) for name in self.keys() if not name.startswith("_")]

    def get(self, name: str) -> SingleTable | None:  # noqa: D102
        return (
            self[name] if not name.startswith("_") and name in self.table_dfs else None
        )

    # Custom operators:

    def __or__(self, other: "DB") -> "DB":  # noqa: D105
        return self.extend(other)

    # Internals:

    def _get_rels(
        self, sources: list[str] | None = None, targets: list[str] | None = None
    ) -> pd.DataFrame:
        """Return all relations contained in this database."""
        return pd.DataFrame.from_records(
            [
                (st, sc, tt, tc)
                for (st, sc), (tt, tc) in self.relations.items()
                if (sources is None or st in sources)
                and (targets is None or tt in targets)
            ],
            columns=["source_table", "source_col", "target_table", "target_col"],
        )

    def _get_valid_refs(
        self, sources: list[str] | None = None, targets: list[str] | None = None
    ) -> pd.DataFrame:
        """Return all valid references contained in this database."""
        rel_vals = []
        rel_keys = []

        all_rels = self._get_rels(sources, targets)

        for t, rels in all_rels.groupby("source_table"):
            table = self[str(t)]
            df = (
                table.df.rename_axis("id", axis="index")
                if not table.df.index.name
                else table.df
            )

            for tt, r in rels.groupby("target_table"):
                f_df = pd.DataFrame(
                    {
                        c: (
                            df[[c]]
                            .reset_index()
                            .merge(
                                self[str(tt)].df.pipe(
                                    lambda df: df.rename_axis("id", axis="index")
                                    if not df.index.name
                                    else df
                                ),
                                left_on=c,
                                right_on=tc,
                            )
                            .set_index(df.index.name)[str(c)]
                            .groupby(df.index.name)
                            .agg("first")
                        )
                        for c, tc in r[["source_col", "target_col"]].itertuples(
                            index=False
                        )
                    }
                )
                rel_vals.append(f_df)
                rel_keys.append((t, tt))

        return pd.concat(rel_vals, keys=rel_keys, names=["src_table", "target_table"])

    def _manifest_virtual_table(
        self, name: str, rel: tuple[str, str] | None = None
    ) -> "SingleTable":
        """Manifest a virtual table with the required cols."""
        rels = (
            self._get_rels(targets=[name])
            if rel is None
            else pd.DataFrame.from_records(
                [(*rel, *self.relations[rel])],
                columns=["source_table", "source_col", "target_table", "target_col"],
            )
        )
        if len(rels) == 0:
            raise ValueError(f"Cannot find any relations with target table '{name}'.")

        frames = []
        for tc, rel_group in rels.groupby("target_col"):
            col_values = reduce(
                set.union,
                (
                    set(self[st][sc].unique())
                    for st, sc in rel_group[["source_table", "source_col"]]
                    .drop_duplicates()
                    .itertuples(index=False)
                ),
            )
            frames.append(pd.DataFrame({tc: list(col_values)}))

        return SingleTable(name=name, db=self, df=pd.concat(frames, ignore_index=True))

    def _isotropic_trim(self) -> "DB":
        """Return new database without orphan data (data w/ no refs to or from)."""
        # Get the status of each single reference.
        valid_refs = self._get_valid_refs()

        result = {}
        for t, table in self.items():
            f = pd.Series(False, index=table.df.index)

            try:
                # Include all rows with any valid outgoing reference.
                outgoing = valid_refs.loc[(t, slice(None), slice(None)), :]
                assert isinstance(outgoing, pd.DataFrame)
                for _, refs in outgoing.groupby("target_table"):
                    f |= (
                        refs.droplevel(["src_table", "target_table"])
                        .notna()
                        .any(axis="columns")
                    )
            except KeyError:
                pass

            try:
                # Include all rows with any valid incoming reference.
                incoming = valid_refs.loc[(slice(None), t, slice(None)), :]
                assert isinstance(incoming, pd.DataFrame)
                for _, refs in incoming.groupby("src_table"):
                    for _, col in refs.items():
                        f |= pd.Series(True, index=col.unique())
            except KeyError:
                pass

            result[t] = table.df.loc[f]

        return DB(
            table_dfs=result,
            relations=self.relations,
            join_tables=self.join_tables,
            schema=self.schema,
        )

    def _centered_trim(
        self, centers: list[str], circuit_breakers: list[str] | None = None
    ) -> "DB":
        """Return new database minus data without (indirect) refs to any given table."""
        circuit_breakers = circuit_breakers or []

        # Get the status of each single reference.
        valid_refs = self._get_valid_refs()

        current_stage = {c: set(self[c].df.index) for c in centers}
        visit_counts = {
            t: pd.Series(0, index=table.df.index) for t, table in self.items()
        }
        visited: set[str] = set()

        while any(len(s) > 0 for s in current_stage.values()):
            next_stage = {t: set() for t in self.keys()}
            for t, idx in current_stage.items():
                if t in visited and t in circuit_breakers:
                    continue

                current, additions = visit_counts[t].align(
                    pd.Series(1, index=list(idx)), fill_value=0
                )
                visit_counts[t] = current + additions

                idx_sel = list(
                    idx & set(visit_counts[t].loc[visit_counts[t] == 1].index)
                )

                if len(idx_sel) == 0:
                    continue

                visited |= set([t])

                try:
                    # Include all rows with any valid outgoing reference.
                    outgoing = valid_refs.loc[(t, slice(None), idx_sel), :]
                    assert isinstance(outgoing, pd.DataFrame)
                    for tt, refs in outgoing.groupby("target_table"):
                        for _, col in refs.items():
                            next_stage[str(tt)] |= set(col.dropna().unique())
                except KeyError:
                    pass

                try:
                    # Include all rows with any valid incoming reference.
                    incoming = valid_refs.loc[(slice(None), t, slice(None)), :]
                    assert isinstance(incoming, pd.DataFrame)
                    for st, refs in incoming.groupby("src_table"):
                        next_stage[str(st)] |= set(
                            refs.droplevel(["src_table", "target_table"])
                            .isin(idx_sel)
                            .any(axis="columns")
                            .replace(False, pd.NA)
                            .dropna()
                            .index
                        )
                except KeyError:
                    pass

            current_stage = next_stage

        return DB(
            table_dfs={t: self[t].df.loc[rc > 0] for t, rc in visit_counts.items()},
            relations=self.relations,
            join_tables=self.join_tables,
            schema=self.schema,
        )



"""Test the geo module."""

import country_converter as coco
import pandas as pd
import pytest

from py_research.geo import (
    CountryScheme,
    GeoRegion,
    GeoScheme,
    countries_to_scheme,
    expand_geo_col_to_cc,
    gen_flag_url,
    match_to_geo_region,
    merge_geo_regions,
)


@pytest.mark.parametrize(
    "countries, target, src, expected",
    [
        (
            pd.Series(["United States", "Canada", "Mexico"]),
            GeoScheme.cc_iso3,
            GeoScheme.country_name,
            pd.Series(["USA", "CAN", "MEX"]),
        ),
        (
            pd.Series(["USA", "CAN", "MEX"]),
            GeoScheme.cc_iso3,
            None,
            pd.Series(["USA", "CAN", "MEX"]),
        ),
        (
            pd.Series([], dtype=object),
            GeoScheme.cc_iso3,
            None,
            pd.Series([], dtype=object),
        ),
    ],
)
def test_countries_to_scheme(
    countries: pd.Series, target: CountryScheme, src: CountryScheme, expected: pd.Series
):
    """Test the countries_to_scheme function."""
    result = countries_to_scheme(countries, target, src)
    pd.testing.assert_series_equal(
        result, expected, check_names=False, check_dtype=False
    )


@pytest.mark.parametrize(
    "df, geo_col, scheme, cc_scheme, expected",
    [
        (
            pd.DataFrame({"geo_region": ["America", "Europe"]}),
            "geo_region",
            GeoScheme.continent,
            GeoScheme.cc_iso3,
            set(
                [
                    *coco.convert(["America"], src="continent"),
                    *coco.convert(["Europe"], src="continent"),
                ]
            ),
        ),
        (
            pd.DataFrame({"geo_region": ["Asia"]}),
            "geo_region",
            GeoScheme.continent,
            GeoScheme.cc_iso3,
            set(
                [
                    *coco.convert(["Asia"], src="continent"),
                ]
            ),
        ),
    ],
)
def test_expand_geo_col_to_cc(
    df: pd.DataFrame,
    geo_col: str,
    scheme: GeoScheme,
    cc_scheme: CountryScheme,
    expected: pd.DataFrame,
):
    """Test the expand_geo_col_to_cc function."""
    result = expand_geo_col_to_cc(df, geo_col, scheme, cc_scheme)
    assert set(result["cc"]) == expected


@pytest.mark.parametrize(
    "df, geo_col, geo_regions, input_scheme, expected",
    [
        (
            pd.DataFrame(
                {"cc": ["USA", "CAN", "MEX", "FRA", "DEU", "GBR", "CHN", "JPN", "IND"]}
            ),
            "cc",
            [
                GeoRegion("DEU", GeoScheme.cc_iso3),
                GeoRegion("America", GeoScheme.continent),
                GeoRegion("EU", GeoScheme.alliance),
            ],
            GeoScheme.cc_iso3,
            pd.Series(
                [
                    "America",
                    "America",
                    "America",
                    "Rest of EU",
                    "Germany",
                    "Rest of world",
                    "Rest of world",
                    "Rest of world",
                    "Rest of world",
                ]
            ).sort_values(),
        ),
        (
            pd.DataFrame({"cc": []}),
            "cc",
            [
                GeoRegion("DEU", GeoScheme.cc_iso3),
                GeoRegion("America", GeoScheme.continent),
                GeoRegion("EU", GeoScheme.alliance),
            ],
            GeoScheme.cc_iso3,
            pd.Series([], dtype=object),
        ),
    ],
)
def test_merge_geo_regions(
    df: pd.DataFrame,
    geo_col: str,
    geo_regions: list[GeoRegion],
    input_scheme: GeoScheme,
    expected: pd.Series,
):
    """Test the merge_geo_regions function."""
    result = merge_geo_regions(df, geo_col, geo_regions, input_scheme)
    pd.testing.assert_series_equal(
        result["geo_region"].sort_values(),
        expected,
        check_index=False,
        check_dtype=False,
        check_names=False,
    )


@pytest.mark.parametrize(
    "countries, geo_region, country_scheme, expected",
    [
        (
            pd.Series(["USA", "CAN", "MEX"]),
            GeoRegion("America", GeoScheme.continent),
            GeoScheme.cc_iso3,
            pd.Series([True, True, True]),
        ),
        (
            pd.Series(["FRA", "DEU", "GBR"]),
            GeoRegion("Europe", GeoScheme.continent),
            GeoScheme.cc_iso3,
            pd.Series([True, True, True]),
        ),
        (
            pd.Series(["CHN", "JPN", "IND"]),
            GeoRegion("Asia", GeoScheme.continent),
            GeoScheme.cc_iso3,
            pd.Series([True, True, True]),
        ),
        (
            pd.Series(["USA", "CAN", "DEU"]),
            GeoRegion("EU", GeoScheme.alliance),
            GeoScheme.cc_iso3,
            pd.Series([False, False, True]),
        ),
    ],
)
def test_match_to_geo_region(
    countries: pd.Series,
    geo_region: GeoRegion,
    country_scheme: CountryScheme,
    expected: pd.Series,
):
    """Test the match_to_geo_region function."""
    result = match_to_geo_region(countries, geo_region, country_scheme)
    pd.testing.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    "cc, width",
    [
        (pd.Series(["USA"]), 160),
        (pd.Series(["FRA"]), 320),
        (pd.Series(["CHN"]), 320),
        (pd.Series(["USA", "FRA", "CHN"]), 160),
    ],
)
def test_gen_flag_url(cc: pd.Series, width: int):
    """Test the gen_flag_url function."""
    result = gen_flag_url(cc, width)
    assert isinstance(result, pd.Series)



"""Helper functions for handling files and directories."""

from pathlib import Path


def ensure_dir_exists(path: Path | str):
    """Make sure the given path exists and is a directory.

    Create dir if necessary. Then return the path unchanged.

    Args:
        path: Path to ensure exists.

    Returns:
        The given path.
    """
    path = Path(path).absolute()
    p = Path(path.parts[0])
    for d in path.parts[1:]:
        p = p / d
        if not p.is_dir():
            if not p.exists():
                p.mkdir()
            else:
                raise NotADirectoryError()

    return path



"""Utilities for working with date and time in data."""

from collections.abc import Callable
from datetime import datetime
from typing import Any, cast

import pandas as pd
from typing_extensions import deprecated

from py_research.data import parse_dtype


def _auto_interval_format(
    time_interval: pd.offsets.BaseOffset, min_bin: datetime, max_bin: datetime
) -> tuple[str | Callable[[Any], str], str]:
    format_func = "%c"
    interval_name = "interval"

    date_prefix = (
        ""
        if max_bin < min_bin + pd.offsets.DateOffset(days=1)
        else "%d "
        if max_bin < min_bin + pd.offsets.DateOffset(months=1)
        else "%m-%d "
        if max_bin < min_bin + pd.offsets.DateOffset(years=1)
        else "%Y-%m-%d "
    )

    match (time_interval):
        case pd.offsets.YearBegin() | pd.offsets.YearEnd():
            format_func = "%Y"
            interval_name = "year"
        case pd.offsets.QuarterBegin() | pd.offsets.QuarterEnd():

            def quarter_format(d: datetime) -> str:
                q = (d.month - 1) // 3 + 1
                if max_bin < min_bin + pd.offsets.DateOffset(years=1):
                    return f"Q{q}"
                return f"{d.year} Q{q}"

            format_func = quarter_format
        case pd.offsets.MonthBegin() | pd.offsets.MonthEnd():
            interval_name = "month"
            if max_bin < min_bin + pd.offsets.DateOffset(years=1):
                format_func = "%Y"
            else:
                format_func = "%Y-%m"
        case pd.offsets.Day():
            interval_name = "day"
            if max_bin < min_bin + pd.offsets.DateOffset(years=1):
                format_func = "%Y"
            elif max_bin < min_bin + pd.offsets.DateOffset(months=1):
                format_func = "%Y-%m"
            else:
                format_func = "%Y-%m-%d"
        case pd.offsets.Week():
            interval_name = "week"
            if max_bin < min_bin + pd.offsets.DateOffset(years=1):
                format_func = "%W"
            else:
                format_func = "%Y week %W"
        case pd.offsets.Hour():
            interval_name = "time"
            format_func = f"{date_prefix}%Hh"
        case pd.offsets.Minute():
            interval_name = "time"
            if max_bin < min_bin + pd.offsets.DateOffset(hours=1):
                format_func = "%M"
            else:
                format_func = f"{date_prefix}%H:%M"
        case pd.offsets.Second():
            interval_name = "time"
            if max_bin < min_bin + pd.offsets.DateOffset(minutes=1):
                format_func = f"{date_prefix}%S"
            elif max_bin < min_bin + pd.offsets.DateOffset(hours=1):
                format_func = f"{date_prefix}%M:%S"
            else:
                format_func = f"{date_prefix}%H:%M:%S"
        case _:
            interval_name = "datetime"
            format_func = "%c"

    return format_func, interval_name


@deprecated("Use datetime formatting options of ``py_research.intl`` instead.")
def datetime_to_interval_series(
    datetime_series: pd.Series,
    time_interval: pd.offsets.BaseOffset = pd.offsets.YearEnd(),
    format: str | None = None,
    interval_col: str | None = None,
) -> pd.Series:
    """Assign intervals matching ``datetime_col`` to new column.

    Args:
        datetime_series: Series of datetime values.
        time_interval: Interval to use for grouping.
        format: Format to use for the interval column.
        interval_col: Name of the interval column.

    Returns:
        Series of intervals matching ``datetime_col``.
    """
    datetime_df = (
        datetime_series.to_frame().assign(time_bin=datetime_series).reset_index()
    )
    resampled = (
        datetime_df.resample(time_interval, group_keys=True, on="time_bin")
        .apply(lambda df: pd.DataFrame(df, index=df.index))
        .reset_index()
    ).set_index(datetime_series.index.name or "index")["time_bin"]

    format_func = format
    if format_func is None:
        format_func, interval_name = _auto_interval_format(
            time_interval, resampled.min(), resampled.max()
        )
        interval_col = interval_col or interval_name

    def apply_format(s: pd.Series) -> pd.Series:
        assert format_func is not None
        return (
            s.dt.strftime(format_func)
            if isinstance(format_func, str)
            else s.map(format_func)
        )

    if time_interval.n == 1:
        resampled = apply_format(resampled)
        resampled = parse_dtype(resampled)
    else:
        start = resampled
        end = cast(pd.Series, resampled + time_interval)  # type: ignore
        resampled = apply_format(start) + " - " + apply_format(end)

        interval_col = (
            f"{interval_col + ' ' if interval_col is not None else ''}interval"
        )

    return resampled.rename(interval_col)



"""Simple & semi-automatic color-system for data visualization purposes."""

import re
from colorsys import hls_to_rgb, rgb_to_hls
from contextvars import ContextVar
from dataclasses import dataclass, field
from os import environ
from pathlib import Path
from typing import Any

from webcolors import IntegerRGB, hex_to_rgb, name_to_rgb, rgb_to_hex
from yaml import CLoader, load


def default_highlights():
    """Return default highlight colors."""
    return {
        0: "#663399",
        1: "#1E90FF",
        2: "#3CB371",
    }


def default_scales():
    """Return default color scale."""
    return {
        0: [
            (0, "#ebf6fa"),
            (0.01, "#d6edf5"),
            (0.03, "#99d3e6"),
            (0.10, "#5cb8d6"),
            (0.30, "#14d5cc"),
            (1, "#35d27e"),
        ],
        1: [
            (0, "#005b7f"),
            (0.15, "#39a9cd"),
            (0.30, "#14d5cc"),
            (1, "#35d27e"),
        ],
    }


@dataclass
class ColorTheme:
    """Define a color theme. Can be used as context manager to activate the theme."""

    highlights: dict[Any, str] = field(default_factory=default_highlights)
    """Highlight colors for different, known categories."""

    scales: dict[Any, list[tuple[float, str]]] = field(default_factory=default_scales)
    """Named or numbered color sclales."""

    def __post_init__(self):  # noqa: D105
        self.__token = None

    def activate(self):
        """Set this as the current theme."""
        self.__token = active_theme.set(self)

    def __enter__(self):  # noqa: D105
        self.activate()
        return self

    def __exit__(self, *_):  # noqa: D105
        if self.__token is not None:
            active_theme.reset(self.__token)


color_file_path = Path(environ.get("COLOR_FILE_PATH") or (Path.cwd() / "colors.yaml"))


def _get_colors_from_file() -> ColorTheme | None:
    if color_file_path.is_file():
        with open(color_file_path, encoding="utf-8") as f:
            return ColorTheme(**load(f, Loader=CLoader))


active_theme: ContextVar[ColorTheme] = ContextVar(
    "active_color_theme", default=(_get_colors_from_file() or ColorTheme())
)


def _parse_css_color(color: str) -> IntegerRGB:
    if (rgb_match := re.match(r"^rgb\(([0-9]+,[0-9]+,[0-9]+)\)$", color)) is not None:
        return IntegerRGB(*(int(val) for val in rgb_match.groups()[0].split(",")))
    elif re.match(r"^#[0-9abcdefABCDEF]{3,6}$", color) is not None:
        return hex_to_rgb(color)
    else:
        try:
            return name_to_rgb(color)
        except ValueError as exc:
            raise ValueError(f"CSS color definition not recognized: {color}") from exc


def _integer_rgb_to_float(color: IntegerRGB) -> tuple[float, float, float]:
    return (color.red / 255, color.green / 255, color.blue / 255)


def _float_to_integer_rgb(color: tuple[float, float, float]) -> IntegerRGB:
    return IntegerRGB(*(int(c * 255) for c in color))


def _adjust_lightness(color: IntegerRGB, lightness: float) -> IntegerRGB:
    hls = rgb_to_hls(*_integer_rgb_to_float(color))
    changed = (hls[0], lightness, hls[2])
    return _float_to_integer_rgb(hls_to_rgb(*changed))


def get_theme() -> ColorTheme:
    """Return active color theme (may be derived from current execution context)."""
    return active_theme.get()


def to_bg_color(color: str, lightness: float = 0.8) -> str:
    """Turn a highlight color into a background color."""
    return rgb_to_hex(_adjust_lightness(_parse_css_color(color), lightness))



"""Documentation-friendly enum classes."""

from strenum import LowercaseStrEnum


class StrEnum(LowercaseStrEnum):
    """StrEnum which renders only its value in a string context."""

    def __repr__(self) -> str:  # noqa: D105
        return f"'{self.value}'"



"""Test telemetry module."""

from py_research.telemetry import get_logger


def test_get_logger():
    """Test get_logger."""
    # Test with no arguments
    logger = get_logger()
    logger.info("test")

    # Test with a name
    logger = get_logger("test")
    logger.info("test")
    assert logger.name == "test"



"""Test tables module."""

import pandas as pd

from py_research.tables import ResultTable, TableStyle, to_html


def test_result_table():
    """Test ResultTable."""
    # Create a DataFrame
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})

    # Create a ResultTable
    table = ResultTable(df)

    # Test the render method
    rendered = to_html(table.to_styled_df())
    assert isinstance(rendered, str)
    assert "A" in rendered
    assert "B" in rendered
    assert "C" in rendered

    # Test the styles attribute
    style = TableStyle(cols=["A"], css={"font-weight": "bold"})
    table.styles.append(style)
    rendered = to_html(table.to_styled_df())
    assert "font-weight: bold;" in rendered

    # Test the labels attribute
    table.labels = {"A": "Alpha", "B": "Beta", "C": "Gamma"}
    rendered = to_html(table.to_styled_df())
    assert "Alpha" in rendered
    assert "Beta" in rendered
    assert "Gamma" in rendered

    # Test widths attribute
    table.widths = {"A": "10rem", "B": "100px", "C": "auto"}
    rendered = to_html(table.to_styled_df())
    assert "width: 10rem;" in rendered
    assert "width: 100px;" in rendered
    assert "width: auto;" in rendered

    # Test relative widths attribute
    table.widths = {"A": 1, "B": 2, "C": 2}
    rendered = to_html(table.to_styled_df())
    assert "width: 20.0%;" in rendered
    assert "width: 40.0%;" in rendered
    assert "width: 40.0%;" in rendered

    # Test the title attribute
    table.title = "Test Table"
    rendered = to_html(table.to_styled_df())
    assert "Test Table" in rendered



"""Test files module."""

from pathlib import Path
from tempfile import gettempdir

import pytest
from py_research.files import ensure_dir_exists


def test_ensure_dir_exists():
    """Test ensure_dir_exists."""
    path = Path(gettempdir()) / "test_dir"
    try:
        assert ensure_dir_exists(path) == path.absolute()
        assert path.is_dir()
    finally:
        path.rmdir()  # Clean up

    # Test with a string
    path_str = f"{gettempdir()}/test_dir"
    try:
        assert ensure_dir_exists(path_str) == Path(path_str).absolute()
        assert Path(path_str).is_dir()
    finally:
        Path(path_str).rmdir()  # Clean up


def test_ensure_dir_exists_not_a_dir():
    """Test ensure_dir_exists with a file."""
    # Create a file
    path = Path(gettempdir()) / "test_file"
    path.touch()
    try:
        with pytest.raises(NotADirectoryError):
            ensure_dir_exists(path)
    finally:
        path.unlink()  # Clean up



"""Utilities for data handling."""

import locale
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from fractions import Fraction
from typing import Any

import numpy.typing as npt
import pandas as pd
from pandas.api.types import (
    is_bool_dtype,
    is_datetime64_any_dtype,
    is_float_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_string_dtype,
    is_timedelta64_dtype,
)
from pandas.core.dtypes.base import ExtensionDtype
from pandas.errors import ParserError

from py_research.hashing import gen_int_hash

YES = ["y", "t", "1", "yes", "true"]
NO = ["n", "f", "0", "no", "false"]


def is_number_dtype(dtype: str | npt.DTypeLike | ExtensionDtype) -> bool:
    """Check if dtype is number-like.

    Args:
      dtype: dtype to check.

    Returns:
      True if dtype is number-like.
    """
    return (
        is_numeric_dtype(dtype)
        or is_datetime64_any_dtype(dtype)
        or is_timedelta64_dtype(dtype)
    )


def to_boolean(s: pd.Series) -> pd.Series:
    """Parse boolean series from string series.

    Args:
      s: string series.

    Returns:
      Boolean series.
    """
    s_lower = s.str.lower()
    if not s_lower.isin([*YES, *NO]).all():
        raise ValueError("Series contains invalid values.")

    return s_lower.isin(YES)


def to_integer(s: pd.Series) -> pd.Series:
    """Parse integer series from string series with locale-awareness.

    Args:
      s: string series.

    Returns:
      Integer series.
    """
    return s.astype(str).map(locale.atoi)


def to_float(s: pd.Series) -> pd.Series:
    """Parse float series from string series with locale-awareness.

    Args:
      s: string series.

    Returns:
      Float series.
    """
    return s.astype(str).map(locale.atof)


def parse_dtype(  # noqa: C901
    s: pd.Series,
    dtype: str | type | npt.DTypeLike | None = None,
    src_locale: str | None = None,
) -> pd.Series:
    """Parse series to dtype with locale-awareness.

    Args:
      s: series to convert.
      dtype: dtype to convert to.
      src_locale: locale to use for conversion.

    Returns:
      Converted series.
    """
    if s.dtype == "object":
        s = s.astype(str)

    if not is_string_dtype(s):
        return s

    result = None

    context_locale = None
    if src_locale is not None:
        try:
            context_locale, _ = locale.getlocale(locale.LC_ALL)
        except (TypeError, ValueError):
            context_locale = None
        locale.setlocale(locale.LC_ALL, src_locale)

    if dtype is not None:
        result = (
            to_boolean(s)
            if is_bool_dtype(dtype)
            else to_integer(s)
            if is_integer_dtype(dtype)
            else to_float(s)
            if is_float_dtype(dtype)
            else s
        ).astype(
            dtype  # type: ignore
        )
    else:
        try:
            result = to_boolean(s)
        except (ValueError, TypeError):
            try:
                result = to_integer(s)
            except (ValueError, TypeError):
                try:
                    result = to_float(s)
                except (ValueError, TypeError):
                    try:
                        result = pd.to_numeric(s)
                    except (ValueError, TypeError):
                        try:
                            result = pd.to_datetime(s, infer_datetime_format=True)
                        except (ParserError, ValueError, TypeError):
                            if s.nunique() < len(s) / 5:
                                result = s.astype("category")

                            if src_locale is None:
                                result = parse_dtype(s, dtype, "C")
                            else:
                                result = s

    if src_locale is not None and context_locale is not None:
        locale.setlocale(locale.LC_ALL, context_locale)

    return result


def gen_id(x: Any, length: int = 10, raw_str: bool = False) -> str:
    """Generate stable hash for obj (must be known, hashable or composed of such).

    Args:
        x: Object to hash.
        length: Length of the hash.
        raw_str:
            Whether to use the raw string representation of the object,
            if it is a string.

    Returns:
        Hash of the object as string.
    """
    s = None
    match (x):
        case int() | float() | complex() | Decimal():
            s = str(x)
        case Fraction():
            s = str(x).replace("/", "_over_")
        case str() if raw_str:
            s = x
        case date():
            s = x.isoformat()
        case time():
            s = f"HHMMSS{'ffffff' if x.microsecond != 0 else ''}".format(x)
        case datetime():
            s = str(x.timestamp())
        case timedelta():
            s = str(x.total_seconds())
        case _:
            s = str(abs(gen_int_hash(x)))

    return s[:length]



"""Connect to an SQL server, validate its schema and perform cached analysis queries."""

from collections.abc import Callable, Mapping
from contextvars import ContextVar
from dataclasses import InitVar, dataclass
from datetime import datetime
from functools import partial, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

import pandas as pd
import sqlalchemy as sqla
import sqlalchemy.orm as orm
from pandas.api.types import (
    is_datetime64_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_string_dtype,
)
from pandas.util import hash_pandas_object
from typing_extensions import Self

from py_research.reflect import get_all_subclasses


def _hash_df(df: pd.DataFrame | pd.Series) -> str:
    return hex(abs(sum(hash_pandas_object(df))))[2:12]


Params = ParamSpec("Params")


def _get_sql_ctx() -> "DB":
    sql_con = active_con.get()

    if sql_con is None:
        raise (RuntimeError("`SQLContext` object not supplied via context."))

    return sql_con


SchemaBase = orm.DeclarativeBase


default_type_map = {
    str: sqla.types.String().with_variant(
        sqla.types.String(50), "oracle"
    ),  # Avoid oracle error when VARCHAR has no size parameter
}


V = TypeVar("V")
S = TypeVar("S", bound=SchemaBase)
S_cov = TypeVar("S_cov", bound=SchemaBase, covariant=True)
S_contra = TypeVar("S_contra", bound=SchemaBase, contravariant=True)
V_contra = TypeVar("V_contra", contravariant=True)


class ColRef(orm.InstrumentedAttribute[V], Generic[V, S_contra]):
    """Reference a column by scheme type, name and value type."""  # type: ignore[override]


class Col(orm.MappedColumn[V]):
    """Define table column within a table schema."""

    if TYPE_CHECKING:

        @overload
        def __get__(self, instance: None, owner: type[S]) -> ColRef[V, S]:
            ...

        @overload
        def __get__(self, instance: object, owner: type) -> V:
            ...

        def __get__(  # noqa: D105
            self, instance: object | None, owner: type[S]
        ) -> ColRef[V, S] | V:
            ...

    @classmethod
    def cast_sqla_attr(cls, attr: orm.MappedColumn) -> Self:
        """Cast a sqlalchemy attr object to this type."""
        return cast(Self, attr)


def _wrap_mapped_col(
    func: Callable[Params, orm.MappedColumn[V]]
) -> Callable[Params, Col[V]]:
    @wraps(func)
    def inner(*args: Params.args, **kwargs: Params.kwargs) -> Col[V]:
        return Col.cast_sqla_attr(func(*args, **kwargs))

    return inner


col = _wrap_mapped_col(orm.mapped_column)
Relation = sqla.ForeignKey


class Data(Protocol[S_cov]):
    """SQL table-shaped data."""

    @property
    def columns(self) -> Mapping[str, sqla.ColumnElement]:
        """Return all columns within this data."""
        ...

    def __clause_element__(self) -> sqla.FromClause:  # noqa: D105
        ...

    def __getitem__(  # noqa: D105
        self, ref: str | ColRef[V, S_cov]
    ) -> sqla.ColumnElement[V]:
        ...

    @property
    def name(self) -> str | None:
        """Data object's name, if any."""
        ...


@dataclass
class Table(Data[S_cov]):
    """Reference to a manifested SQL table."""

    sqla_table: sqla.Table

    @property
    def columns(self) -> dict[str, sqla.Column]:  # noqa: D102
        return dict(self.__clause_element__().columns)

    def __getitem__(self, ref: str | ColRef[V, S_cov]) -> sqla.Column[V]:  # noqa: D105
        return self.columns[ref if isinstance(ref, str) else ref.key]

    def __clause_element__(self) -> sqla.Table:  # noqa: D105
        return self.sqla_table

    @property
    def name(self) -> str:  # noqa: D102
        return self.sqla_table.name


def _map_df_dtype(c: pd.Series) -> sqla.types.TypeEngine:
    if is_datetime64_dtype(c):
        return sqla.types.DATETIME()
    elif is_integer_dtype(c):
        return sqla.types.INTEGER()
    elif is_numeric_dtype(c):
        return sqla.types.FLOAT()
    elif is_string_dtype(c):
        max_len = c.str.len().max()
        if max_len < 16:
            return sqla.types.CHAR(max_len)
        elif max_len < 256:
            return sqla.types.VARCHAR(max_len)

    return sqla.types.BLOB()


def _cols_from_df(df: pd.DataFrame) -> dict[str, sqla.Column]:
    if len(df.index.names) > 1:
        raise NotImplementedError("Multi-index not supported yet.")

    return {
        **{
            level: sqla.Column(
                level,
                _map_df_dtype(df.index.get_level_values(level).to_series()),
                primary_key=True,
            )
            for level in df.index.names
        },
        **{
            str(name): sqla.Column(str(col.name), _map_df_dtype(col))
            for name, col in df.items()
        },
    }


@dataclass
class Query(Data[S_cov]):
    """SQL data defined by a query."""

    sel: sqla.Select
    """SQLAlchemy select statement."""

    query_name: InitVar[str]
    """Name of this query."""

    schema: type[S_cov] | None = None
    """Schema of this query's result."""

    _subquery: sqla.Subquery | None = None

    def __post_init__(self, query_name: str):  # noqa: D105
        self.__name = query_name

    @property
    def name(self) -> str:
        """Name of this query."""
        return self.__name

    @property
    def columns(self) -> dict[str, sqla.ColumnElement]:  # noqa: D102
        return dict(self.__clause_element__().columns)

    def __clause_element__(self) -> sqla.Subquery:  # noqa: D105
        if self._subquery is None:
            self._subquery = self.sel.subquery()

        return self._subquery

    def __getitem__(  # noqa: D105
        self, ref: str | ColRef[V, S_cov]
    ) -> sqla.ColumnElement[V]:
        return self.columns[ref if isinstance(ref, str) else ref.key]


SelFunc: TypeAlias = Callable[Params, sqla.Select | sqla.Table]
BoundSelFunc: TypeAlias = Callable[[], sqla.Select | sqla.Table]


@dataclass
class DeferredQuery(Data[S_cov]):
    """SQL data defined by a query or table-returning Python function."""

    func: BoundSelFunc
    """Function returning a SQL query or table."""

    schema: type[S_cov] | None = None
    """Schema of this query's result."""

    _subquery: sqla.Subquery | sqla.Table | None = None

    @property
    def columns(self) -> dict[str, sqla.ColumnElement]:  # noqa: D102
        return dict(self.__clause_element__().columns)

    def __clause_element__(self) -> sqla.Subquery | sqla.Table:  # noqa: D105
        if self._subquery is None:
            res = self.func()
            self._subquery = res.subquery() if isinstance(res, sqla.Select) else res

        return self._subquery

    def __getitem__(  # noqa: D105
        self, ref: str | ColRef[V, S_cov]
    ) -> sqla.ColumnElement[V]:
        return self.columns[ref if isinstance(ref, str) else ref.key]

    @property
    def name(self) -> str | None:  # noqa: D102
        return self.__clause_element__().name or (
            self.func.func.__name__
            if isinstance(self.func, partial)
            else self.func.__name__
            if isinstance(self.func, Callable)
            else None
        )


DS = TypeVar("DS", bound="DBSchema")


def _map_foreignkey_schema(
    table: sqla.Table,
    to_schema: str | None,
    constraint: sqla.ForeignKeyConstraint,
    referred_schema: str | None,
    schema_dict: "dict[str | None, Schema | None]",
) -> str | None:
    assert to_schema in schema_dict

    for schema_name, schema in schema_dict.items():
        if schema is not None and schema.schema_def is not None:
            for schema_class in get_all_subclasses(schema.schema_def):
                if (
                    hasattr(schema_class, "__table__")
                    and schema_class.__table__ is constraint.referred_table
                ):
                    return schema_name

    return referred_schema


@dataclass
class Schema(Generic[S_cov, DS]):
    """SQL schema defining multiple related tables."""

    schema_def: type[S_cov] | None = None
    """Schema base class for this schema's tables."""

    db_schema: type[DS] | None = None
    """DBSchema instance this schema is defined in."""

    default: bool = False
    """Whether this is the default schema for its schema type."""

    name: str | None = None
    """Name of this schema."""

    def __set_name__(self, _: type, name: str):  # noqa: D105
        self.name = name if name != "__default__" else None

    def __get__(  # noqa: D105
        self, instance: Any, owner: type[DS]
    ) -> "Schema[S_cov, DS]":
        self.db_schema = owner
        return self

    @property
    def tables(self) -> dict[str, Table]:
        """Tables defined by this schema."""
        assert self.db_schema is not None
        return {
            name: Table(table)
            for name, table in self.db_schema.metadata().tables.items()
            if table.schema == self.name
        }

    def __getitem__(self, table: type[S] | str) -> Table[S]:  # noqa: D105
        # TODO: Find a way to limit table to type[S_cov] without losing type info,
        # maybe via decorating the schema classes.
        assert self.db_schema is not None
        return Table(
            self.db_schema.metadata().tables[
                ".".join(
                    [
                        *([self.name] if self.name else []),
                        table if isinstance(table, str) else table.__tablename__,
                    ]
                )
            ]
        )


class DBSchema(Generic[S_cov]):
    """Schema for an entire SQL database."""

    __default__: Schema[S_cov, Self] = Schema()
    """Default schema for this DB."""

    @classmethod
    def schema_dict(cls) -> dict[str | None, Schema | None]:
        """Return dict with all sub-schemas."""
        return {
            None: cls.__default__,  # type: ignore
            **{
                (name if name != "__default__" else None): s
                for name, s in cls.__dict__.items()
                if isinstance(s, Schema) and name != "__default__"
            },
        }

    @classmethod
    def default(cls) -> Schema[S_cov, Self]:
        """Return default schema."""
        return cls.__default__  # type: ignore

    @classmethod
    def defaults(cls) -> dict[type[SchemaBase], Schema]:
        """Return mapping of schema types to default schema names."""
        defaults = {}
        for s in cls.schema_dict().values():
            if (
                s is not None
                and s.schema_def is not None
                and (s.schema_def not in defaults.keys() or s.default)
            ):
                defaults[s.schema_def] = s
        return defaults

    @classmethod
    def metadata(cls) -> sqla.MetaData:
        """Return metadata object containing all this DB's schemas and tables."""
        if hasattr(cls, "__metadata__"):
            return cls.__metadata__

        cls.__metadata__ = sqla.MetaData()
        schema_dict = cls.schema_dict()
        for schema in schema_dict.values():
            if schema is not None and schema.schema_def is not None:
                for table in schema.schema_def.metadata.tables.values():
                    table.to_metadata(
                        cls.__metadata__,
                        schema=schema.name,
                        referred_schema_fn=partial(
                            _map_foreignkey_schema, schema_dict=schema_dict
                        ),
                    )

        return cls.__metadata__


S2 = TypeVar("S2", bound=SchemaBase)


def _remove_external_fk(table: sqla.Table):
    """Dirty vodoo to remove external FKs from existing table."""
    for c in table.columns.values():
        c.foreign_keys = set(
            fk
            for fk in c.foreign_keys
            if fk.constraint and fk.constraint.referred_table.schema == table.schema
        )

    table.foreign_keys = set(  # type: ignore
        fk
        for fk in table.foreign_keys
        if fk.constraint and fk.constraint.referred_table.schema == table.schema
    )

    table.constraints = set(
        c
        for c in table.constraints
        if not isinstance(c, sqla.ForeignKeyConstraint)
        or c.referred_table.schema == table.schema
    )


@dataclass
class DB(Generic[S_cov, DS]):
    """Active connection to a SQL server."""

    url: str | sqla.URL
    """Connection URL."""

    schema: type[DBSchema[S_cov]] = DBSchema
    """Schema for this DB."""

    tag: str = datetime.today().strftime("YYYY-MM-dd")
    """Tag for this DB session. Used to identify automatically created tables."""

    validate: bool = True
    """Whether to validate the DB schema on connection."""

    def __post_init__(self):  # noqa: D105
        self.__token = None

        self.engine = sqla.create_engine(self.url)

        if self.validate:
            # Validation.

            inspector = sqla.inspect(self.engine)

            for table in self.schema.metadata().tables.values():
                if table.schema is not None:
                    assert inspector.has_schema(table.schema)

                assert inspector.has_table(table.name, table.schema)

                db_columns = {
                    c["name"]: c
                    for c in inspector.get_columns(table.name, table.schema)
                }
                for column in table.columns:
                    assert column.name in db_columns

                    db_col = db_columns[column.name]
                    assert isinstance(db_col["type"], type(column.type))
                    assert (
                        db_col["nullable"] == column.nullable or column.nullable is None
                    )

                db_pk = inspector.get_pk_constraint(table.name, table.schema)
                if (
                    len(db_pk["constrained_columns"]) > 0
                ):  # Allow source tbales without pk
                    assert set(db_pk["constrained_columns"]) == set(
                        table.primary_key.columns.keys()
                    )

                db_fks = inspector.get_foreign_keys(table.name, table.schema)
                for fk in table.foreign_key_constraints:
                    matches = [
                        (
                            set(db_fk["constrained_columns"]) == set(fk.column_keys),
                            (
                                db_fk["referred_table"].lower()
                                == fk.referred_table.name.lower()
                            ),
                            set(db_fk["referred_columns"])
                            == set(f.column.name for f in fk.elements),
                        )
                        for db_fk in db_fks
                    ]

                    assert any(all(m) for m in matches)

    @overload
    def __getitem__(self, key: Schema[S2, DS]) -> Schema[S2, DS]:
        ...

    @overload
    def __getitem__(self, key: None) -> Schema[S_cov, DS]:
        ...

    def __getitem__(  # noqa: D105
        self, key: Schema[S2, DS] | None
    ) -> Schema[S2, DS] | Schema[S_cov, DS]:
        return self.schema.default() if key is None else key  # type: ignore

    def activate(self) -> None:
        """Set this instance as the current SQL context."""
        self.__token = active_con.set(self)

    def __enter__(self):  # noqa: D105
        self.activate()
        return self

    def __exit__(self, *_):  # noqa: D105
        if self.__token is not None:
            active_con.reset(self.__token)

    def default_table(self, schema: type[S]) -> Table[S]:
        """Return default table for given table schema.

        Args:
            schema: Schema of the table to return.

        Returns:
            Default table for given schema.
        """
        defaults = self.schema.defaults()

        valid_keys = list(filter(lambda s: issubclass(schema, s), defaults.keys()))
        if len(valid_keys) == 0:
            raise KeyError(schema)

        schema_name = defaults[valid_keys[0]].name
        return Table(
            self.schema.metadata().tables[
                ".".join(
                    [
                        *([schema_name] if schema_name else []),
                        schema.__tablename__,
                    ]
                )
            ]
        )

    def _table_from_df(
        self,
        df: pd.DataFrame,
        schema: type[S_cov] | None = None,
        name: str | None = None,
    ) -> Table[S_cov]:
        table_name = f"{self.tag}_df_{name + '_' if name else ''}{_hash_df(df)}"

        if table_name not in self.schema.metadata().tables and not sqla.inspect(
            self.engine
        ).has_table(table_name, schema=None):
            cols = (
                {
                    c.key: sqla.Column(c.key, c.type, primary_key=c.primary_key)
                    for c in schema.__table__.columns
                }
                if schema is not None
                else _cols_from_df(df)
            )
            table = Table(
                sqla.Table(table_name, self.schema.metadata(), *cols.values())
            )
            table.sqla_table.create(self.engine)
            df.reset_index()[list(cols.keys())].to_sql(
                table_name, self.engine, if_exists="append", index=False
            )

        return (
            Table(self.schema.metadata().tables[table_name])
            if table_name in self.schema.metadata().tables
            else Table(
                sqla.Table(
                    table_name, self.schema.metadata(), autoload_with=self.engine
                )
            )
        )

    def _table_from_query(
        self, q: Query[S] | DeferredQuery[S], with_external_fk: bool = False
    ) -> Table[S_cov]:
        table_name = f"{self.tag}_query_{q.name}"

        if table_name in self.schema.metadata().tables:
            return Table(self.schema.metadata().tables[table_name])
        elif sqla.inspect(self.engine).has_table(table_name, schema=None):
            return Table(
                sqla.Table(
                    table_name,
                    self.schema.metadata(),
                    autoload_with=self.engine,
                )
            )
        else:
            sel_res = q.__clause_element__()

            sqla_table = (
                q.schema.__table__.to_metadata(
                    self.schema.metadata(),
                    schema=None,  # type: ignore
                    referred_schema_fn=partial(
                        _map_foreignkey_schema, schema_dict=self.schema.schema_dict()
                    ),
                    name=table_name,
                )
                if q.schema is not None and isinstance(q.schema.__table__, sqla.Table)
                else sqla.Table(
                    table_name,
                    self.schema.metadata(),
                    *(
                        sqla.Column(name, col.type, primary_key=col.primary_key)
                        for name, col in sel_res.columns.items()
                    ),
                )
            )
            if not with_external_fk:
                _remove_external_fk(sqla_table)
            sqla_table.create(self.engine)

            table = Table(sqla_table)

            with self.engine.begin() as con:
                con.execute(
                    sqla.insert(table).from_select(
                        sel_res.exported_columns.keys(), sel_res
                    )
                )

            return table

    def to_table(
        self,
        src: pd.DataFrame | Query[S_cov] | DeferredQuery[S_cov],
        schema: type[S_cov] | None = None,
        name: str | None = None,
        with_external_fk: bool = False,
    ) -> Table[S_cov]:
        """Transfer dataframe or sql query results to manifested SQL table.

        Args:
            src: Source / definition of data.
            schema: Schema of the table to create.
            name: Name of the table to create.
            with_external_fk: Whether to include foreign keys to external tables.

        Returns:
            Reference to manifested SQL table.
        """
        match src:
            case pd.DataFrame():
                return self._table_from_df(src, schema, name)
            case Query() | DeferredQuery():
                return self._table_from_query(src, with_external_fk)

    def to_df(self, q: Data) -> pd.DataFrame:
        """Transfer manifested table or query results to local dataframe.

        Args:
            q: Source / definition of data.

        Returns:
            Dataframe containing the data.
        """
        sel = sqla.select(q)
        with self.engine.connect() as con:
            return pd.read_sql(sel, con)


active_con: ContextVar[DB | None] = ContextVar("active_sql_con", default=None)


QueryFunc: TypeAlias = Callable[Params, Query[S_cov]]
DefQueryFunc: TypeAlias = Callable[Params, DeferredQuery[S_cov]]


@overload
def query(func: SelFunc[Params]) -> QueryFunc[Params, S]:
    ...


@overload
def query(
    *, defer: Literal[True] = ..., schema: type[S] = ...  # type: ignore
) -> Callable[[SelFunc[Params]], DefQueryFunc[Params, S]]:
    ...


@overload
def query(
    *, defer: Literal[True] = ..., schema: None = ...
) -> Callable[[SelFunc[Params]], DefQueryFunc[Params, Any]]:
    ...


@overload
def query(
    *, defer: Literal[False] = ..., schema: type[S] = ...  # type: ignore
) -> Callable[[SelFunc[Params]], QueryFunc[Params, S]]:
    ...


@overload
def query(
    *, defer: Literal[False] = ..., schema: None = ...
) -> Callable[[SelFunc[Params]], QueryFunc[Params, Any]]:
    ...


def query(
    func: SelFunc[Params] | None = None,
    *,
    defer: bool | None = False,
    schema: type[S] | None = None,
) -> (
    QueryFunc[Params, S]
    | Callable[[SelFunc[Params]], QueryFunc[Params, S] | DefQueryFunc[Params, S]]
):
    """Decorator to transform query-returning function into a referencable query.

    Args:
        func: Function returning a :py:obj:`sqla.Selectable`.
        defer: Whether to defer execution of the query.
        schema: Schema of the query's result.

    Returns:
        Decorated function.
    """

    def inner(
        func: SelFunc[Params],
    ) -> QueryFunc[Params, S] | DefQueryFunc[Params, S]:
        @wraps(func)
        def inner_inner(
            *args: Params.args, **kwargs: Params.kwargs
        ) -> Query[S] | DeferredQuery[S]:
            res = func(*args, **kwargs)
            return (
                Query(
                    res if isinstance(res, sqla.Select) else sqla.select(res),
                    func.__name__,
                    schema,
                )
                if not defer
                else DeferredQuery(partial(func, *args, **kwargs), schema)
            )

        return inner_inner  # type: ignore

    return inner(func) if func is not None else inner  # type: ignore


def default_table(schema: type[S]) -> DeferredQuery[S]:
    """Return default table for given table schema.

    Args:
        schema: Schema of the table to return.

    Returns:
        Reference to default table for given schema.
    """

    def get_current_default_table() -> sqla.Table:
        ctx = _get_sql_ctx()
        return ctx.default_table(schema).sqla_table

    return DeferredQuery(get_current_default_table, schema)



"""Test the db module."""

from datetime import date
from pathlib import Path
from tempfile import gettempdir

import pandas as pd
import pytest
from py_research.data import parse_dtype
from py_research.db import DB, DBSchema, Table


@pytest.fixture
def table_df_projects():
    """Return a dataframe for the projects table."""
    return (
        pd.DataFrame(
            {
                "id": [1, 2, 3, 4],
                "name": [
                    "baking cake",
                    "cleaning shoes",
                    "fixing cars",
                    "writing book",
                ],
                "start": [
                    date(2020, 1, 1),
                    date(2020, 1, 2),
                    date(2020, 1, 3),
                    date(2020, 1, 4),
                ],
                "end": [
                    date(2020, 1, 4),
                    date(2020, 1, 5),
                    date(2020, 1, 6),
                    date(2020, 1, 7),
                ],
                "status": ["done", "done", "started", "abandoned"],
            }
        )
        .set_index("id")
        .apply(parse_dtype, axis="index")
    )


@pytest.fixture
def table_df_persons():
    """Return a dataframe for the persons table."""
    return (
        pd.DataFrame(
            {
                "id": ["a", "b", "c"],
                "name": ["John", "Jane", "Joe"],
                "age": [20, 30, 40],
                "height": [1.8, 1.7, 1.9],
                "weight": [80, 70, 90],
                "gender": ["male", "male", "diverse"],
            }
        )
        .set_index("id")
        .apply(parse_dtype, axis="index")
    )


@pytest.fixture
def table_df_memberships():
    """Return a dataframe for the memberships table."""
    return (
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "project": [1, 2, 3, 1],
                "member": ["a", "b", "c", "b"],
                "role": ["leader", "member", "member", "member"],
            }
        )
        .set_index(["id"])
        .apply(parse_dtype, axis="index")
    )


@pytest.fixture
def table_df_tasks():
    """Return a dataframe for the tasks table."""
    return (
        pd.DataFrame(
            {
                "id": [1, 2, 3, 4, 5, 6, 7],
                "name": ["task1", "task2", "task3", "task4", "task5", "task6", "task7"],
                "project": [1, 1, 1, 2, 2, 3, 3],
                "assignee": ["a", "a", "b", "a", "b", "a", "c"],
                "status": ["todo", "todo", "done", "todo", "todo", "todo", "done"],
            }
        )
        .set_index("id")
        .apply(parse_dtype, axis="index")
    )


@pytest.fixture
def relations():
    """Return a dict of relations between table columns."""
    return {
        ("memberships", "project"): ("projects", "id"),
        ("memberships", "member"): ("persons", "id"),
        ("tasks", "project"): ("projects", "id"),
        ("tasks", "assignee"): ("persons", "id"),
        ("persons", "gender"): ("genders", "name"),
    }


@pytest.fixture
def join_tables():
    """Return a list of join tables."""
    return set(["memberships"])


@pytest.fixture
def db_from_tables(
    table_df_projects: pd.DataFrame,
    table_df_persons: pd.DataFrame,
    table_df_memberships: pd.DataFrame,
    table_df_tasks: pd.DataFrame,
    relations: dict[tuple[str, str], tuple[str, str]],
    join_tables: set[str],
):
    """Return DB instance with tables."""
    db = DB(
        {
            "projects": table_df_projects,
            "persons": table_df_persons,
            "memberships": table_df_memberships,
            "tasks": table_df_tasks,
        },
        relations,
        join_tables,
    )
    return db


def test_create_db_empty():
    """Test the creation of a DB instance."""
    db = DB()
    assert isinstance(db, DB)


def test_create_db_from_tables(db_from_tables: DB):
    """Test the creation of a DB instance from tables."""
    assert isinstance(db_from_tables, DB)
    assert set(db_from_tables.keys()) == {
        "projects",
        "persons",
        "memberships",
        "tasks",
        "genders",
    }
    assert isinstance(db_from_tables["projects"].df, pd.DataFrame)
    assert "persons" in db_from_tables
    assert "genders" in db_from_tables
    assert set(db_from_tables["genders"]["name"]) == {"male", "diverse"}
    assert len(db_from_tables) == 5


def test_trim_db_anisotropic(db_from_tables: DB):
    """Test the trimming of a DB instance."""
    db = db_from_tables.trim()
    assert isinstance(db, DB)
    assert set(db["projects"].df.index) == {1, 2, 3}


def test_extend_db(db_from_tables: DB):
    """Test the extension of a DB instance."""
    db = db_from_tables.extend(
        {
            "projects": pd.DataFrame(
                {
                    "id": [5, 6],
                    "name": ["baking cake", "cleaning shoes"],
                    "start": [date(2020, 1, 1), date(2020, 1, 2)],
                    "end": [date(2020, 1, 4), date(2020, 1, 5)],
                    "status": ["done", "done"],
                }
            ).set_index("id"),
            "persons": pd.DataFrame(
                {
                    "id": ["d", "e"],
                    "name": ["John", "Jane"],
                    "age": [20, 30],
                    "height": [1.8, 1.7],
                    "weight": [80, 70],
                }
            ).set_index("id"),
            "memberships": pd.DataFrame(
                {
                    "id": [4, 5],
                    "project": ["a", "b"],
                    "member": [1, 2],
                    "role": ["leader", "member"],
                }
            ).set_index(["id"]),
            "tasks": pd.DataFrame(
                {
                    "id": [8, 9],
                    "name": ["task8", "task9"],
                    "project": [1, 1],
                    "assignee": ["a", "b"],
                    "status": ["todo", "todo"],
                }
            ).set_index("id"),
        }
    )
    assert isinstance(db, DB)
    assert set(db["projects"].df.index) == {1, 2, 3, 4, 5, 6}
    assert set(db["persons"].df.index) == {"a", "b", "c", "d", "e"}
    assert set(db["memberships"].df.index) == {0, 1, 2, 3, 4, 5}
    assert set(db["tasks"].df.index) == {1, 2, 3, 4, 5, 6, 7, 8, 9}


def test_filter_db(db_from_tables: DB):
    """Test the filtering of a DB instance."""
    db = db_from_tables.filter(
        {"projects": db_from_tables["projects"].df["status"] == "done"}
    )
    assert isinstance(db, DB)
    assert set(db["projects"].df.index) == {1, 2}
    assert set(db["persons"].df.index) == {"a", "b"}
    assert len(db["memberships"].df) == 3
    assert len(db["tasks"].df) == 6


def test_save_load_db(db_from_tables: DB):
    """Test the saving of a DB instance."""
    tempdir = Path(gettempdir())
    db_file_path = tempdir / "test_db.xlsx"

    db_from_tables.schema = DBSchema  # Use base class as empty schema.
    db_from_tables.save(db_file_path)
    assert db_file_path.is_file()

    db_loaded = DB.load(db_file_path)
    assert isinstance(db_loaded, DB)
    for t in db_loaded.keys():
        pd.testing.assert_frame_equal(db_from_tables[t].df, db_loaded[t].df)
    assert db_from_tables.relations == db_loaded.relations
    assert db_from_tables.join_tables == db_loaded.join_tables
    assert db_from_tables.updates == db_loaded.updates
    assert db_from_tables.schema == db_loaded.schema


def test_table_to_from_excel(db_from_tables: DB):
    """Test saving / loading table to / from excel."""
    tempdir = Path(gettempdir())
    table_file_path = tempdir / "test_table.xlsx"

    db_from_tables["persons"].to_excel(table_file_path)
    loaded_persons = Table.from_excel(table_file_path)
    pd.testing.assert_frame_equal(db_from_tables["persons"].df, loaded_persons.df)
    assert db_from_tables["persons"].source_map == loaded_persons.source_map

    merged = db_from_tables["persons"].merge(right=db_from_tables["projects"])
    merged.to_excel(table_file_path)
    merged_loaded = Table.from_excel(table_file_path)
    pd.testing.assert_frame_equal(merged.df, merged_loaded.df)
    assert merged.source_map == merged_loaded.source_map


def test_filter_db_table(db_from_tables: DB):
    """Test the filtering of a DB table."""
    table = db_from_tables["projects"]
    filtered_table = table.filter(table.df["status"] == "done")
    assert isinstance(filtered_table, Table)
    assert set(filtered_table.df.index) == {1, 2}


def test_set_db_table(db_from_tables: DB):
    """Test the filtering of a DB table."""
    table = db_from_tables["projects"]
    filtered_table = table.filter(table.df["status"] == "done")

    db_from_tables["projects"] = filtered_table
    assert set(db_from_tables["projects"].df.index) == {1, 2}
    with pytest.raises(ValueError):
        db_from_tables["projects"] = filtered_table.df.reset_index()[["name"]]


def test_merge_db_table_forward(db_from_tables: DB):
    """Test the forward merging of a DB table."""
    table = db_from_tables["tasks"]

    table_merged = table.merge(link_to_right="project", naming="source")
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten(".", "always")
    assert set(db_from_tables["tasks"]["name"].unique()) == set(
        table_flat["tasks.name"].unique()
    )
    assert set(db_from_tables["tasks"]["project"].unique()) == set(
        table_flat["projects.id"].unique()
    )


def test_merge_db_table_forward_virtual(db_from_tables: DB):
    """Test the forward merging of a DB table."""
    table = db_from_tables["persons"]

    table_merged = table.merge(link_to_right="gender", naming="source")
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten(".", "always")
    assert set(db_from_tables["persons"]["name"].unique()) == set(
        table_flat["persons.name"].unique()
    )
    assert set(db_from_tables["persons"]["gender"].unique()) == set(
        table_flat["genders.name"].unique()
    )


def test_merge_db_table_forward_filtered(db_from_tables: DB):
    """Test the forward merging of a DB table with filters."""
    left_table = db_from_tables["tasks"].filter(
        db_from_tables["tasks"]["status"] == "todo"
    )
    right_table = db_from_tables["projects"].filter(
        db_from_tables["projects"]["status"] == "done"
    )

    table_merged = left_table.merge(
        right=right_table, link_to_right="project", naming="path"
    )
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(left_table["name"].unique()) == set(table_flat["tasks->name"].unique())
    assert set(right_table.df.index) == set(
        table_flat["tasks->project->id"].dropna().unique()
    )


def test_merge_db_table_backward(db_from_tables: DB):
    """Test the backward merging of a DB table."""
    table = db_from_tables["projects"]

    table_merged = table.merge(right=db_from_tables["tasks"], naming="path")
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(db_from_tables["projects"]["name"].unique()) == set(
        table_flat["projects->name"].unique()
    )
    assert set(db_from_tables["tasks"]["project"].unique()) == set(
        table_flat["projects<=tasks->project"].dropna().unique()
    )


def test_merge_db_table_backward_explicit(db_from_tables: DB):
    """Test the explicit backward merging of a DB table."""
    table = db_from_tables["projects"]

    table_merged = table.merge(
        right=db_from_tables["tasks"], link_to_left="project", naming="path"
    )
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(db_from_tables["projects"]["name"].unique()) == set(
        table_flat["projects->name"].unique()
    )
    assert set(db_from_tables["tasks"]["project"].unique()) == set(
        table_flat["projects<=tasks->project"].dropna().unique()
    )


def test_merge_db_table_double(db_from_tables: DB):
    """Test the double merging of a DB table."""
    table = db_from_tables["persons"]

    table_merged = table.merge(right=db_from_tables["projects"], naming="path")
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(db_from_tables["persons"]["name"].unique()) == set(
        table_flat["persons->name"].unique()
    )
    assert set(db_from_tables["memberships"]["project"].unique()) == set(
        table_flat["persons->project->id"].unique()
    )


def test_merge_db_table_double_filtered(db_from_tables: DB):
    """Test the double merging of a DB table with filtering."""
    left_table = db_from_tables["persons"].filter(db_from_tables["persons"]["age"] < 30)
    link_table = db_from_tables["memberships"].filter(
        db_from_tables["memberships"]["role"] == "leader"
    )

    table_merged = left_table.merge(
        right=db_from_tables["projects"], link_table=link_table, naming="path"
    )
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(left_table["name"].unique()) == set(table_flat["persons->name"].unique())
    assert set(table_flat["persons->project->id"].unique()) == {1}


def test_merge_db_table_double_indefinite(db_from_tables: DB):
    """Test the indefinite double merging of a DB table."""
    table = db_from_tables["persons"]

    table_merged = table.merge(link_table=db_from_tables["memberships"], naming="path")
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(db_from_tables["persons"]["name"].unique()) == set(
        table_flat["persons->name"].unique()
    )
    assert set(db_from_tables["memberships"]["project"].unique()) == set(
        table_flat["persons->project->id"].unique()
    )


def test_merge_db_table_double_explicit(db_from_tables: DB):
    """Test the indefinite double merging of a DB table."""
    table = db_from_tables["persons"]

    table_merged = table.merge(
        link_table=db_from_tables["memberships"],
        right=db_from_tables["projects"],
        naming="path",
    )
    assert isinstance(table_merged, Table)
    assert table_merged.df.columns.nlevels == 2

    table_flat = table_merged.flatten("->", "always")
    assert set(db_from_tables["persons"]["name"].unique()) == set(
        table_flat["persons->name"].unique()
    )
    assert set(db_from_tables["memberships"]["project"].unique()) == set(
        table_flat["persons->project->id"].unique()
    )


def test_extract_db_table(db_from_tables: DB):
    """Test the extraction of a db table."""
    table = db_from_tables["projects"]

    table_merged = table.merge(right=db_from_tables["tasks"], naming="path")

    extracted_db1 = table_merged.extract()
    assert isinstance(extracted_db1, DB)
    assert set(extracted_db1.keys()) == {
        "projects",
        "tasks",
        "memberships",
        "persons",
        "genders",
    }
    assert set(extracted_db1["projects"].df.index) == {1, 2, 3, 4}

    extracted_db2 = table_merged.extract(with_relations=False)
    assert isinstance(extracted_db2, DB)
    assert set(extracted_db2.keys()) == {"projects", "tasks"}
    assert set(extracted_db2["projects"].df.index) == {1, 2, 3, 4}


def test_extend_db_table(db_from_tables: DB):
    """Test the extending of a DB table."""
    table = db_from_tables["projects"]

    table_extended = table.extend(
        pd.DataFrame(
            {
                "id": [5, 6],
                "name": ["baking cake", "cleaning shoes"],
                "start": [date(2020, 1, 1), date(2020, 1, 2)],
                "end": [date(2020, 1, 4), date(2020, 1, 5)],
                "status": ["done", "done"],
            }
        ).set_index("id")
    )
    assert isinstance(table_extended, Table)
    assert set(table_extended.df.index) == {1, 2, 3, 4, 5, 6}


def test_db_to_graph(db_from_tables: DB):
    """Test transformation of database to graph."""
    nodes, edges = db_from_tables.to_graph(
        [
            trimmed_proj := db_from_tables["projects"].trim(["name", "status"]),
            trimmed_pers := db_from_tables["persons"].trim(["name", "age"]),
            "tasks",
        ]
    )
    assert isinstance(nodes, pd.DataFrame)
    assert len(nodes) == len(trimmed_proj.df) + len(trimmed_pers.df) + len(
        db_from_tables["tasks"].df
    )
    assert set(nodes.columns) == {
        "node_id",
        "table",
        "id",
        "name",
        "status",
        "project",
        "assignee",
        "age",
    }
    assert isinstance(edges, pd.DataFrame)
    assert set(edges.columns) == {"source", "target", "ltr", "rtl", "role"}


def test_db_to_graph_merged(db_from_tables: DB):
    """Test transformation of database to graph."""
    table = db_from_tables["persons"]
    table_merged = table.merge(right=db_from_tables["projects"], naming="path")

    nodes, edges = db_from_tables.to_graph([table_merged])
    assert isinstance(nodes, pd.DataFrame)
    assert set(nodes.columns) == {
        "node_id",
        "table",
        "id",
        "name",
        "status",
        "age",
        "start",
        "end",
        "height",
        "weight",
        "gender",
    }
    assert isinstance(edges, pd.DataFrame)
    assert set(edges.columns) == {"source", "target", "ltr", "rtl", "role"}



"""Functions for ranking entities by multiple criteria."""
from typing import Literal

import numpy as np
import pandas as pd

from py_research.intl import get_localization

RankMode = Literal["ascending", "descending"]


def create_rank_series(
    rank_by: pd.Series | list[pd.Series] | pd.DataFrame,
    rank_mode: RankMode = "descending",
    exclude: pd.Series | None = None,
    name: str = "rank",
) -> pd.Series:
    """Create a series of ranks given a series of values to rank by.

    Args:
        rank_by: Series of values to rank by.
        rank_mode: Whether to rank ascending or descending.
        exclude: Values to exclude from ranking.
        name: Name of the resulting series.

    Returns:
        Series of ranks.
    """
    data_table = pd.DataFrame(rank_by)
    filtered_data = data_table.loc[~exclude] if exclude is not None else data_table

    ranks = pd.Series(
        np.arange(1, len(filtered_data) + 1),
        index=filtered_data.sort_values(
            by=list(filtered_data.columns), ascending=(rank_mode == "ascending")
        ).index,
        name=name,
    )

    if ranks.empty:
        return ranks

    rank_by_cols = list(filtered_data.columns)
    for _, g in filtered_data.groupby(
        by=(rank_by_cols[0] if len(rank_by_cols) == 1 else rank_by_cols)
    ):
        ranks.loc[g.index] = min(ranks.loc[g.index])

    return data_table.join(ranks, how="left")[name]


def create_ranking_filter(
    rank_by: pd.Series,
    cutoff_rank: int,
    sort_order: Literal["ascending", "descending"] = "descending",
    rank_only: pd.Series | None = None,
    show_only: pd.Series | None = None,
    show_always: pd.Series | None = None,
    count_always_shown: bool = False,
) -> pd.Series:
    """Create a filter based on ranking.

    Args:
        rank_by: Series of values to rank by.
        cutoff_rank: Cutoff rank for the filter.
        sort_order: Whether to rank ascending or descending.
        rank_only: Only rank the given values.
        show_only: Only show the given values.
        show_always: Always show the given values.
        count_always_shown: Whether to count the values in show_always as ranks.

    Returns:
        Boolean series representing the filter. The series name contains a
        description of the filter.
    """
    loc = get_localization()

    def _filter_explanation(
        cutoff: int,
        rank_by: str,
        sort_order: str,
        rank_only: str | None = None,
        show_only: str | None = None,
        show_always: str | None = None,
    ) -> str:
        return "; ".join(
            s
            for s in [
                (
                    f"{cutoff} highest-ranked"
                    + (
                        f" {rank_only}"
                        if rank_only is not None and len(rank_only) <= 20
                        else ""
                    )
                    + f" according to '{rank_by}'"
                    + f" ({sort_order})"
                    + (
                        f", only ranking {rank_only}"
                        if rank_only is not None and len(rank_only) > 20
                        else ""
                    )
                ),
                (f"only showing {show_only}" if show_only is not None else ""),
                (f"always showing {show_always}" if show_always is not None else ""),
            ]
            if s
        )

    desc = loc.text(
        _filter_explanation(
            cutoff_rank,
            str(rank_by.name),
            sort_order,
            rank_only=str(rank_only.name)
            if rank_only is not None and rank_only.name is not None
            else None,
            show_only=str(show_only.name)
            if show_only is not None and show_only.name is not None
            else None,
            show_always=str(show_always.name)
            if show_always is not None and show_always.name is not None
            else None,
        )
    )

    rank_by_filter = np.full(len(rank_by), True)

    if rank_only is not None:
        # Apply a custom filter, if given
        rank_by_filter &= rank_only

    rank_series = create_rank_series(
        rank_by.loc[rank_by_filter],
        sort_order,
        exclude=(show_always if not count_always_shown else None),
    )
    rank_by_filter &= rank_series <= cutoff_rank

    if show_always is not None:
        # Apply a custom filter, if given
        rank_by_filter |= show_always

    if show_only is not None:
        # Apply a custom filter, if given
        rank_by_filter &= show_only

    return pd.Series(rank_by_filter, index=rank_by.index, name=desc)



"""Test plots module."""

from pathlib import Path
from tempfile import gettempdir

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from py_research.plots import export_html, with_dropdown


def test_export_html():
    """Test export_html."""
    # Create a simple figure
    fig = go.Figure(data=[{"type": "scatter", "x": [1, 2, 3], "y": [4, 5, 6]}])

    # Test with a Path object
    path = Path(gettempdir()) / "test_file.html"
    try:
        export_html(fig, path)
        assert path.is_file()
    finally:
        path.unlink()  # Clean up


def plot_func(df: pd.DataFrame) -> go.Figure:
    """Plot a scatter plot of GDP per capita vs population with a slider for year."""
    return px.scatter(
        df,
        x="pop",
        y="gdpPercap",
        color="country",
        size="lifeExp",
        animation_frame="year",
    )


def test_with_dropdown():
    """Test ``with_dropdown`` decorator.""" ""
    df = px.data.gapminder()
    plot_func_wd = with_dropdown("continent")(plot_func)

    fig = plot_func_wd(df)
    assert isinstance(fig, go.Figure)
